import logging
import select
import socketserver
from typing import Optional

from paramiko.transport import Transport


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(socketserver.BaseRequestHandler):
    ssh_transport: Optional[Transport] = None
    chain_host: Optional[str] = None
    chain_port: Optional[int] = None
    logger = logging.Logger(__name__)

    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            self.logger.warning('Incoming request to {host}:{port} failed: {error}'.format(
                host=self.chain_host, port=self.chain_port, error=repr(e)))
            return

        if chan is None:
            self.logger.warning('Incoming request to {host}:{port} was rejected by the SSH server'.format(
                host=self.chain_host, port=self.chain_port))
            return

        self.logger.info('Connected! Tunnel open {} -> {} -> {}'.format(
            self.request.getpeername(),
            chan.getpeername(),
            (self.chain_host, self.chain_port),
        ))

        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peer = self.request.getpeername()
        chan.close()
        self.request.close()
        self.logger.info('Tunnel closed from {}'.format(peer))


def forward_tunnel(local_port, remote_host, remote_port, transport, bind_addr='') -> ForwardServer:
    class SubHandler(Handler):
        ssh_transport = transport
        chain_host = remote_host
        chain_port = remote_port

    return ForwardServer((bind_addr, local_port), SubHandler)


# vim:sw=4:ts=4:et:



import os
import queue
import random
from typing import List, Optional

import requests
import threading
import time

from platypush.context import get_bus
from platypush.plugins import Plugin, action
from platypush.message.event.torrent import \
    TorrentDownloadStartEvent, TorrentDownloadedMetadataEvent, TorrentStateChangeEvent, \
    TorrentDownloadProgressEvent, TorrentDownloadCompletedEvent, TorrentDownloadStopEvent, \
    TorrentPausedEvent, TorrentResumedEvent, TorrentQueuedEvent


class TorrentPlugin(Plugin):
    """
    Plugin to search and download torrents.

    Requires:

        * **python-libtorrent-bin** (``pip install python-libtorrent-bin``)

    """

    # Wait time in seconds between two torrent transfer checks
    _MONITOR_CHECK_INTERVAL = 3

    default_torrent_ports = [6881, 6891]
    categories = {
        'movies': None,
        'tv': None,
    }

    torrent_state = {}
    transfers = {}
    # noinspection HttpUrlsUsage
    default_popcorn_base_url = 'http://popcorn-time.ga'

    def __init__(self, download_dir=None, torrent_ports=None, imdb_key=None, popcorn_base_url=default_popcorn_base_url,
                 **kwargs):
        """
        :param download_dir: Directory where the videos/torrents will be downloaded (default: none)
        :type download_dir: str

        :param torrent_ports: Torrent ports to listen on (default: 6881 and 6891)
        :type torrent_ports: list[int]

        :param imdb_key: The IMDb API (https://imdb-api.com/) is used to search for movies and series. Insert your
            IMDb API key if you want support for content search.
        :type imdb_key: str

        :param popcorn_base_url: Custom base URL to use for the PopcornTime API.
        :type popcorn_base_url: str
        """

        super().__init__(**kwargs)

        if torrent_ports is None:
            torrent_ports = []

        for category in self.categories.keys():
            self.categories[category] = getattr(self, 'search_' + category)

        self.imdb_key = imdb_key
        self.imdb_urls = {}
        self.torrent_ports = torrent_ports if torrent_ports else self.default_torrent_ports
        self.download_dir = None
        self._sessions = {}
        self._lt_session = None
        self.popcorn_base_url = popcorn_base_url
        self.torrent_base_urls = {
            'movies': f'{popcorn_base_url}/movie/{{}}',
            'tv': f'{popcorn_base_url}/show/{{}}',
        }

        if imdb_key:
            self.imdb_urls = {
                'movies': f'https://imdb-api.com/API/SearchMovie/{self.imdb_key}/{{}}',
                'tv': f'https://imdb-api.com/API/SearchSeries/{self.imdb_key}/{{}}',
            }

        if download_dir:
            self.download_dir = os.path.abspath(os.path.expanduser(download_dir))

    @action
    def search(self, query, category=None, language=None, *args, **kwargs):
        """
        Perform a search of video torrents.

        :param query: Query string, video name or partial name
        :type query: str

        :param category: Category to search. Supported types: "movies", "tv".
            Default: None (search all categories)
        :type category: str or list

        :param language: Language code for the results - example: "en" (default: None, no filter)
        :type language: str
        """

        assert self.imdb_key, 'No imdb_key specified'
        results = []
        if isinstance(category, str):
            category = [category]

        # noinspection PyCallingNonCallable
        def worker(cat):
            if cat not in self.categories:
                raise RuntimeError('Unsupported category {}. Supported category: {}'.
                                   format(cat, self.categories.keys()))

            self.logger.info('Searching {} torrents for "{}"'.format(cat, query))
            results.extend(self.categories[cat](query, language=language, *args, **kwargs))

        workers = [
            threading.Thread(target=worker, kwargs={'cat': category})
            for category in (category or self.categories.keys())
        ]

        for worker in workers:
            worker.start()
        for worker in workers:
            worker.join()

        return results

    def _imdb_query(self, query: str, category: str):
        assert self.imdb_key, 'No imdb_key specified'
        imdb_url = self.imdb_urls.get(category)

        assert imdb_url, f'No such category: {category}'
        imdb_url = imdb_url.format(query)
        response = requests.get(imdb_url).json()

        assert not response.get('errorMessage'), response['errorMessage']
        return response.get('results', [])

    def _torrent_search_worker(self, imdb_id: str, category: str, q: queue.Queue):
        base_url = self.torrent_base_urls.get(category)
        assert base_url, f'No such category: {category}'
        try:
            results = requests.get(base_url.format(imdb_id)).json()
            q.put(results)
        except Exception as e:
            q.put(e)

    def _search_torrents(self, query, category):
        imdb_results = self._imdb_query(query, category)
        result_queues = [queue.Queue()] * len(imdb_results)
        workers = [
            threading.Thread(target=self._torrent_search_worker, kwargs={
                'imdb_id': imdb_results[i]['id'],
                'category': category,
                'q': result_queues[i],
            })
            for i in range(len(imdb_results))
        ]

        results = []
        errors = []

        for worker in workers:
            worker.start()
        for q in result_queues:
            res_ = q.get()
            if isinstance(res_, Exception):
                errors.append(res_)
            else:
                results.append(res_)
        for worker in workers:
            worker.join()

        if errors:
            self.logger.warning(f'Torrent search errors: {[str(e) for e in errors]}')

        return results

    @staticmethod
    def _results_to_movies_response(results: List[dict], language: Optional[str] = None):
        return sorted([
            {
                'imdb_id': result.get('imdb_id'),
                'type': 'movies',
                'file': item.get('file'),
                'title': '{title} [movies][{language}][{quality}]'.format(
                    title=result.get('title'), language=lang, quality=quality),
                'duration': int(result.get('runtime'), 0),
                'year': int(result.get('year'), 0),
                'synopsis': result.get('synopsis'),
                'trailer': result.get('trailer'),
                'genres': result.get('genres', []),
                'images': result.get('images', []),
                'rating': result.get('rating', {}),
                'language': lang,
                'quality': quality,
                'size': item.get('size'),
                'provider': item.get('provider'),
                'seeds': item.get('seed'),
                'peers': item.get('peer'),
                'url': item.get('url'),
            }
            for result in results
            for (lang, items) in (result.get('torrents', {}) or {}).items()
            if not language or language == lang
            for (quality, item) in items.items()
            if quality != '0'
        ], key=lambda item: item.get('seeds', 0), reverse=True)

    @staticmethod
    def _results_to_tv_response(results: List[dict]):
        return sorted([
            {
                'imdb_id': result.get('imdb_id'),
                'tvdb_id': result.get('tvdb_id'),
                'type': 'tv',
                'file': item.get('file'),
                'series': result.get('title'),
                'title': '{series} [S{season:02d}E{episode:02d}] {title} [tv][{quality}]'.format(
                    series=result.get('title'),
                    season=episode.get('season'),
                    episode=episode.get('episode'),
                    title=episode.get('title'),
                    quality=quality),
                'duration': int(result.get('runtime'), 0),
                'year': int(result.get('year'), 0),
                'synopsis': result.get('synopsis'),
                'overview': episode.get('overview'),
                'season': episode.get('season'),
                'episode': episode.get('episode'),
                'num_seasons': result.get('num_seasons'),
                'country': result.get('country'),
                'network': result.get('network'),
                'status': result.get('status'),
                'genres': result.get('genres', []),
                'images': result.get('images', []),
                'rating': result.get('rating', {}),
                'quality': quality,
                'provider': item.get('provider'),
                'seeds': item.get('seeds'),
                'peers': item.get('peers'),
                'url': item.get('url'),
            }
            for result in results
            for episode in result.get('episodes', [])
            for quality, item in (episode.get('torrents', {}) or {}).items()
            if quality != '0'
        ], key=lambda item: '{series}.{quality}.{season:02d}.{episode:02d}'.format(
            series=item.get('series'), quality=item.get('quality'),
            season=item.get('season'), episode=item.get('episode')))

    def search_movies(self, query, language=None):
        return self._results_to_movies_response(
            self._search_torrents(query, 'movies'), language=language)

    def search_tv(self, query, **_):
        return self._results_to_tv_response(
            self._search_torrents(query, 'tv'))

    def _get_torrent_info(self, torrent, download_dir):
        import libtorrent as lt

        torrent_file = None
        magnet = None
        info = {}
        file_info = {}

        # noinspection HttpUrlsUsage
        if torrent.startswith('magnet:?'):
            magnet = torrent
            magnet_info = lt.parse_magnet_uri(magnet)
            if isinstance(magnet_info, dict):
                info = {
                    'name': magnet_info.get('name'),
                    'url': magnet,
                    'magnet': magnet,
                    'trackers': magnet_info.get('trackers', []),
                    'save_path': download_dir,
                }
            else:
                info = {
                    'name': magnet_info.name,
                    'url': magnet,
                    'magnet': magnet,
                    'trackers': magnet_info.trackers,
                    'save_path': download_dir,
                }
        elif torrent.startswith('http://') or torrent.startswith('https://'):
            response = requests.get(torrent, allow_redirects=True)
            torrent_file = os.path.join(download_dir, self._generate_rand_filename())

            with open(torrent_file, 'wb') as f:
                f.write(response.content)
        else:
            torrent_file = os.path.abspath(os.path.expanduser(torrent))
            if not os.path.isfile(torrent_file):
                raise RuntimeError('{} is not a valid torrent file'.format(torrent_file))

        if torrent_file:
            file_info = lt.torrent_info(torrent_file)
            # noinspection PyArgumentList
            info = {
                'name': file_info.name(),
                'url': torrent,
                'trackers': [t.url for t in list(file_info.trackers())],
                'save_path': download_dir,
            }

        return info, file_info, torrent_file, magnet

    def _fire_event(self, event, event_hndl):
        bus = get_bus()
        bus.post(event)

        try:
            if event_hndl:
                event_hndl(event)
        except Exception as e:
            self.logger.warning('Exception in torrent event handler: {}'.format(str(e)))
            self.logger.exception(e)

    def _torrent_monitor(self, torrent, transfer, download_dir, event_hndl, is_media):
        def thread():
            files = []
            last_status = None
            download_started = False
            metadata_downloaded = False

            while not transfer.is_finished():
                if torrent not in self.transfers:
                    self.logger.info('Torrent {} has been stopped and removed'.format(torrent))
                    self._fire_event(TorrentDownloadStopEvent(url=torrent), event_hndl)
                    break

                status = transfer.status()
                torrent_file = transfer.torrent_file()

                if torrent_file:
                    self.torrent_state[torrent]['size'] = torrent_file.total_size()
                    files = [os.path.join(
                        download_dir,
                        torrent_file.files().file_path(i))
                        for i in range(0, torrent_file.files().num_files())
                    ]

                    if is_media:
                        from platypush.plugins.media import MediaPlugin
                        # noinspection PyProtectedMember
                        files = [f for f in files if MediaPlugin.is_video_file(f)]

                self.torrent_state[torrent]['download_rate'] = status.download_rate
                self.torrent_state[torrent]['name'] = status.name
                self.torrent_state[torrent]['num_peers'] = status.num_peers
                self.torrent_state[torrent]['paused'] = status.paused
                self.torrent_state[torrent]['progress'] = round(100 * status.progress, 2)
                self.torrent_state[torrent]['state'] = status.state.name
                self.torrent_state[torrent]['title'] = status.name
                self.torrent_state[torrent]['torrent'] = torrent
                self.torrent_state[torrent]['upload_rate'] = status.upload_rate
                self.torrent_state[torrent]['url'] = torrent
                self.torrent_state[torrent]['files'] = files

                if transfer.has_metadata() and not metadata_downloaded:
                    self._fire_event(TorrentDownloadedMetadataEvent(**self.torrent_state[torrent]), event_hndl)
                    metadata_downloaded = True

                if status.state == status.downloading and not download_started:
                    self._fire_event(TorrentDownloadStartEvent(**self.torrent_state[torrent]), event_hndl)
                    download_started = True

                if last_status and status.progress != last_status.progress:
                    self._fire_event(TorrentDownloadProgressEvent(**self.torrent_state[torrent]), event_hndl)

                if not last_status or status.state != last_status.state:
                    self._fire_event(TorrentStateChangeEvent(**self.torrent_state[torrent]), event_hndl)

                if last_status and status.paused != last_status.paused:
                    if status.paused:
                        self._fire_event(TorrentPausedEvent(**self.torrent_state[torrent]), event_hndl)
                    else:
                        self._fire_event(TorrentResumedEvent(**self.torrent_state[torrent]), event_hndl)

                last_status = status
                time.sleep(self._MONITOR_CHECK_INTERVAL)

            if transfer and transfer.is_finished():
                self._fire_event(TorrentDownloadCompletedEvent(**self.torrent_state[torrent]), event_hndl)

            self.remove(torrent)
            return files

        return thread

    def _get_session(self):
        if self._lt_session:
            return self._lt_session

        import libtorrent as lt
        # noinspection PyArgumentList
        self._lt_session = lt.session()
        return self._lt_session

    @action
    def download(self, torrent, download_dir=None, _async=False, event_hndl=None, is_media=False):
        """
        Download a torrent.

        :param torrent: Torrent to download. Supported formats:

            * Magnet URLs
            * Torrent URLs
            * Local torrent file

        :type torrent: str

        :param download_dir: Directory to download, overrides the default download_dir attribute (default: None)
        :type download_dir: str

        :param _async: If true then the method will add the torrent to the transfer and then return. Updates on
            the download status should be retrieved either by listening to torrent events or registering the
            event handler. If false (default) then the method will exit only when the torrent download is complete.
        :type _async: bool

        :param event_hndl: A function that takes an event object as argument and is invoked upon a new torrent event
            (download started, progressing, completed etc.)
        :type event_hndl: function

        :param is_media: Set it to true if you're downloading a media file that you'd like to stream as soon as the
            first chunks are available. If so, then the events and the status method will only include media files
        :type is_media: bool
        """

        if torrent in self.torrent_state and torrent in self.transfers:
            return self.torrent_state[torrent]

        import libtorrent as lt

        if not download_dir:
            if self.download_dir:
                download_dir = self.download_dir
            else:
                raise RuntimeError('download_dir not specified')

        download_dir = os.path.abspath(os.path.expanduser(download_dir))
        os.makedirs(download_dir, exist_ok=True)
        info, file_info, torrent_file, magnet = self._get_torrent_info(torrent, download_dir)

        if torrent in self._sessions:
            self.logger.info('A torrent session is already running for {}'.format(torrent))
            return self.torrent_state.get(torrent, {})

        session = self._get_session()
        session.listen_on(*self.torrent_ports)
        self._sessions[torrent] = session

        params = {
            'save_path': download_dir,
            'storage_mode': lt.storage_mode_t.storage_mode_sparse,
        }

        if magnet:
            transfer = lt.add_magnet_uri(session, magnet, params)
        else:
            params['ti'] = file_info
            transfer = session.add_torrent(params)

        self.transfers[torrent] = transfer
        self.torrent_state[torrent] = {
            'url': torrent,
            'title': transfer.status().name,
            'trackers': info['trackers'],
            'save_path': download_dir,
        }

        self._fire_event(TorrentQueuedEvent(url=torrent), event_hndl)
        self.logger.info('Downloading "{}" to "{}" from [{}]'.format(info['name'], download_dir, torrent))
        monitor_thread = self._torrent_monitor(torrent=torrent, transfer=transfer, download_dir=download_dir,
                                               event_hndl=event_hndl, is_media=is_media)

        if not _async:
            return monitor_thread()

        threading.Thread(target=monitor_thread).start()
        return self.torrent_state[torrent]

    @action
    def status(self, torrent=None):
        """
        Get the status of the current transfers.

        :param torrent: Torrent path, URL or magnet URI whose status will be retrieve (default: None, retrieve all
            current transfers)
        :type torrent: str

        :returns: A dictionary in the format torrent_url -> status
        """

        if torrent:
            return self.torrent_state.get(torrent)
        return self.torrent_state

    @action
    def pause(self, torrent):
        """
        Pause/resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        if self.torrent_state[torrent].get('paused', False):
            self.transfers[torrent].resume()
        else:
            self.transfers[torrent].pause()

        return self.torrent_state[torrent]

    @action
    def resume(self, torrent):
        """
        Resume a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        self.transfers[torrent].resume()

    @action
    def remove(self, torrent):
        """
        Stops and removes a torrent transfer.

        :param torrent: Torrent URL as returned from `get_status()`
        :type torrent: str
        """

        if torrent not in self.transfers:
            return None, "No transfer in progress for {}".format(torrent)

        self.transfers[torrent].pause()
        del self.torrent_state[torrent]
        del self.transfers[torrent]

        if torrent in self._sessions:
            del self._sessions[torrent]

    @action
    def quit(self):
        """
        Quits all the transfers and the active session
        """
        transfers = self.transfers.copy()
        for torrent in transfers:
            self.remove(torrent)

    @staticmethod
    def _generate_rand_filename(length=16):
        name = ''
        for i in range(0, length):
            name += hex(random.randint(0, 15))[2:].upper()
        return name + '.torrent'


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from subprocess import Popen
from typing import Tuple

from platypush.plugins.camera import CameraInfo, Camera


@dataclass
class FFmpegCameraInfo(CameraInfo):
    ffmpeg_args: Tuple[str] = ()

    def to_dict(self) -> dict:
        return {
            'ffmpeg_args': list(self.ffmpeg_args or ()),
            **super().to_dict()
        }


class FFmpegCamera(Camera):
    info: FFmpegCameraInfo
    object: Popen


# vim:sw=4:ts=4:et:



import logging
import re
import requests
import time

from frozendict import frozendict
from threading import Thread

from platypush.message.event.http import HttpEvent
from platypush.utils import set_thread_name


class HttpRequest(object):
    poll_seconds = 60
    timeout = 5

    class HttpRequestArguments(object):
        def __init__(self, url, method='get', *args, **kwargs):
            self.method = method.lower()
            self.url = url
            self.args = args
            self.kwargs = kwargs

    def __init__(self, args, bus=None, poll_seconds=None, timeout=None,
                 skip_first_call=True, **kwargs):
        super().__init__()

        self.poll_seconds = poll_seconds or self.poll_seconds
        self.timeout = timeout or self.timeout
        self.bus = bus
        self.skip_first_call = skip_first_call
        self.last_request_timestamp = 0
        self.logger = logging.getLogger('platypush')

        if isinstance(args, self.HttpRequestArguments):
            self.args = args
        elif isinstance(args, dict):
            self.args = self.HttpRequestArguments(**args)
        else:
            raise RuntimeError('{} is neither a dictionary nor an HttpRequest')

        if 'timeout' not in self.args.kwargs:
            self.args.kwargs['timeout'] = self.timeout

        self.request_args = {
            'method': self.args.method, 'url': self.args.url, **self.args.kwargs
        }

    def execute(self):
        def _thread_func():
            set_thread_name('HttpPoll')
            is_first_call = self.last_request_timestamp == 0
            self.last_request_timestamp = time.time()

            try:
                method = getattr(requests, self.args.method.lower())
                response = method(self.args.url, *self.args.args, **self.args.kwargs)
                new_items = self.get_new_items(response)

                if isinstance(new_items, HttpEvent):
                    event = new_items
                    new_items = event.args['response']
                else:
                    event = HttpEvent(dict(self), new_items)

                if new_items and self.bus:
                    if not self.skip_first_call or (
                            self.skip_first_call and not is_first_call):
                        self.bus.post(event)

                response.raise_for_status()
            except Exception as e:
                self.logger.exception(e)
                self.logger.warning('Encountered an error while retrieving {}: {}'.
                                    format(self.args.url, str(e)))

        Thread(target=_thread_func, name='HttpPoll').start()

    def get_new_items(self, response):
        """ Gets new items out of a response """
        raise NotImplementedError("get_new_items must be implemented in a derived class")

    def __iter__(self):
        for (key, value) in self.request_args.items():
            yield key, value


class JsonHttpRequest(HttpRequest):
    def __init__(self, path=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path
        self.seen_entries = set()

    def get_new_items(self, response):
        response = response.json()
        new_entries = []

        if self.path:
            m = re.match(r'\${\s*(.*)\s*}', self.path)
            response = eval(m.group(1))

        for entry in response:
            flattened_entry = deep_freeze(entry)
            if flattened_entry not in self.seen_entries:
                new_entries.append(entry)
            self.seen_entries.add(flattened_entry)

        return new_entries


def deep_freeze(x):
    if isinstance(x, str) or not hasattr(x, "__len__"):
        return x
    if hasattr(x, "keys") and hasattr(x, "values"):
        return frozendict({deep_freeze(k): deep_freeze(v) for k, v in x.items()})
    if hasattr(x, "__getitem__"):
        return tuple(map(deep_freeze, x))

    return frozenset(map(deep_freeze, x))

# vim:sw=4:ts=4:et:



import logging
import os
import pytest
import time
from threading import Thread

from platypush import Daemon, Config

from .utils import config_file, set_base_url

app_start_timeout = 5


def clear_loggers():
    """
    Remove handlers from all loggers at teardown.
    This is to prevent pytest spitting out logging errors on teardown if the logging objects have been deinitialized
    (see https://github.com/pytest-dev/pytest/issues/5502#issuecomment-647157873).
    """
    import logging
    # noinspection PyUnresolvedReferences
    loggers = [logging.getLogger()] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, 'handlers', [])
        for handler in handlers:
            logger.removeHandler(handler)


@pytest.fixture(scope='session', autouse=True)
def app():
    logging.info('Starting Platypush test service')

    Config.init(config_file)
    app = Daemon(config_file=config_file, redis_queue='platypush-tests/bus')
    Thread(target=lambda: app.run()).start()
    logging.info('Sleeping {} seconds while waiting for the daemon to start up'.format(app_start_timeout))
    time.sleep(app_start_timeout)
    yield app

    logging.info('Stopping Platypush test service')
    app.stop_app()
    clear_loggers()
    db_file = (Config.get('main.db') or {}).get('engine', '')[len('sqlite:///'):]

    if db_file and os.path.isfile(db_file):
        logging.info('Removing temporary db file {}'.format(db_file))
        os.unlink(db_file)


@pytest.fixture(scope='session')
def db_file():
    yield Config.get('main.db')['engine'][len('sqlite:///'):]


@pytest.fixture(scope='session')
def base_url():
    backends = Config.get_backends()
    assert 'http' in backends, 'Missing HTTP server configuration'
    url = 'http://localhost:{port}'.format(port=backends['http']['port'])
    set_base_url(url)
    yield url



import json

from platypush import Config
from platypush.message import Message
from platypush.plugins import Plugin, action


class ConfigPlugin(Plugin):
    """
    This plugin can be used to programmatically access the application configuration.
    """

    @action
    def get(self) -> dict:
        """
        Get the current configuration.
        """
        return Config.get()

    @action
    def get_plugins(self) -> dict:
        """
        Get the configured plugins.
        """
        return Config.get_plugins()

    @action
    def get_backends(self) -> dict:
        """
        Get the configured backends.
        """
        return Config.get_backends()

    @action
    def get_procedures(self) -> dict:
        """
        Get the configured procedures.
        """
        return json.loads(json.dumps(Config.get_procedures(), cls=Message.Encoder))

    @action
    def dashboards(self) -> dict:
        """
        Get the configured dashboards.
        """
        return Config.get_dashboards()

    @action
    def get_dashboard(self, name: str) -> str:
        """
        Get a dashboard configuration by name.
        """
        return Config.get_dashboard(name)

    @action
    def get_device_id(self) -> str:
        """
        Get the configured ``device_id``.
        """
        return Config.get('device_id')


# vim:sw=4:ts=4:et:



from marshmallow import fields, INCLUDE
from marshmallow.schema import Schema

from platypush.schemas import StrippedString


class SlackMessageBlockSchema(Schema):
    class Meta:
        unknown = INCLUDE

    type = fields.String(required=True, metadata=dict(description='Message block type'))
    block_id = fields.String(required=True, metadata=dict(description='Block ID'))


class SlackMessageIconSchema(Schema):
    image_36 = fields.URL()
    image_48 = fields.URL()
    image_72 = fields.URL()


class SlackMessageSchema(Schema):
    text = StrippedString(required=True, metadata=dict(description='Message text'))
    user = fields.String(required=True, metadata=dict(description='User ID of the sender'))
    channel = fields.String(metadata=dict(description='Channel ID associated with the message'))
    team = fields.String(metadata=dict(description='Team ID associated with the message'))
    timestamp = fields.DateTime(metadata=dict(description='Date and time of the event'))
    icons = fields.Nested(SlackMessageIconSchema)
    blocks = fields.Nested(SlackMessageBlockSchema, many=True)
    previous_message = fields.Nested(
        'SlackMessageSchema', metadata=dict(
            description='For received replies, it includes the parent message in the reply chain. '
                        'For edited messages, it contains the previous version.'
        )
    )



import pytest

from platypush.event.hook import EventCondition
from platypush.message.event.ping import PingEvent


condition = EventCondition.build({
    'type': 'platypush.message.event.ping.PingEvent',
    'message': 'This is (the)? answer: ${answer}'
})


def test_event_parse():
    """
    Test for the events/conditions matching logic.
    """
    message = "GARBAGE GARBAGE this is the answer: 42"
    event = PingEvent(message=message)
    result = event.matches_condition(condition)
    assert result.is_match
    assert 'answer' in result.parsed_args
    assert result.parsed_args['answer'] == '42'

    message = "what is not the answer? 43"
    event = PingEvent(message=message)
    result = event.matches_condition(condition)
    assert not result.is_match


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



from marshmallow import fields
from marshmallow.schema import Schema


class SunEventsSchema(Schema):
    sunrise = fields.DateTime(metadata=dict(description='Next sunrise time'))
    sunset = fields.DateTime(metadata=dict(description='Next sunset time'))
    solar_noon = fields.DateTime(metadata=dict(description='Next solar noon time'))
    civil_twilight_begin = fields.DateTime(metadata=dict(description='Next civil twilight start time'))
    civil_twilight_end = fields.DateTime(metadata=dict(description='Next civil twilight end time'))
    nautical_twilight_begin = fields.DateTime(metadata=dict(description='Next nautical twilight start time'))
    nautical_twilight_end = fields.DateTime(metadata=dict(description='Next nautical twilight end time'))
    astronomical_twilight_begin = fields.DateTime(metadata=dict(description='Next astronomical twilight start time'))
    astronomical_twilight_end = fields.DateTime(metadata=dict(description='Next astronomical twilight end time'))



from marshmallow import fields
from marshmallow.schema import Schema


class MediaCollectionSchema(Schema):
    id = fields.String(
        metadata=dict(
            description='Collection ID',
        )
    )

    name = fields.String(
        required=True,
        metadata=dict(
            description='Collection name',
        )
    )

    type = fields.String(
        metadata=dict(
            description='Collection type (movies, music, series etc.)',
        )
    )

    image = fields.URL(
        metadata=dict(
            description='Collection image (URL)',
        )
    )


class MediaArtistSchema(Schema):
    id = fields.String(
        metadata=dict(
            description='Artist ID',
        )
    )

    name = fields.String(
        required=True,
        metadata=dict(
            description='Artist name',
        )
    )

    image = fields.URL(
        metadata=dict(
            description='Artist main image (URL)',
        )
    )


class MediaItemSchema(Schema):
    id = fields.String()
    title = fields.String(required=True)
    url = fields.URL()
    file = fields.String()
    image = fields.URL()


class MediaVideoSchema(MediaItemSchema):
    year = fields.Integer()
    has_subtitles = fields.Boolean()


class MediaMovieSchema(MediaItemSchema):
    pass


class MediaEpisodeSchema(MediaItemSchema):
    pass




from abc import ABCMeta
from typing import Optional, Iterable, Union

from platypush.message.event import Event


class JoystickEvent(Event):
    """
    Generic joystick event.
    """

    def __init__(self, code, state, *args, **kwargs):
        """
        :param code: Event code, usually the code of the source key/handle
        :type code: str

        :param state: State of the triggering element. Can be 0/1 for a button, -1/0/1 for an axis, a discrete integer
            for an analog input etc.
        :type state: int
        """

        super().__init__(*args, code=code, state=state, **kwargs)


class _JoystickEvent(Event, ABCMeta):
    """
    Base joystick event class.
    """

    def __init__(self, device: str, **kwargs):
        super().__init__(device=device, **kwargs)


class JoystickConnectedEvent(_JoystickEvent):
    """
    Event triggered upon joystick connection.
    """
    def __init__(self,
                 *args,
                 name: Optional[str] = None,
                 axes: Optional[Iterable[Union[int, str]]] = None,
                 buttons: Optional[Iterable[Union[int, str]]] = None,
                 **kwargs):
        """
        :param name: Device name.
        :param axes: List of the device axes, as indices or names.
        :param buttons: List of the device buttons, as indices or names.
        """
        super().__init__(*args, name=name, axes=axes, buttons=buttons, **kwargs)


class JoystickDisconnectedEvent(_JoystickEvent):
    """
    Event triggered upon joystick disconnection.
    """


class JoystickStateEvent(_JoystickEvent):
    """
    Event triggered when the state of the joystick changes.
    """

    def __init__(self, *args, axes: Iterable[int], buttons: Iterable[bool], **kwargs):
        """
        :param axes: Joystick axes values, as a list of integer values.
        :param buttons: Joystick buttons values, as a list of boolean values (True for pressed, False for released).
        """
        super().__init__(*args, axes=axes, buttons=buttons, **kwargs)


class JoystickButtonPressedEvent(_JoystickEvent):
    """
    Event triggered when a joystick button is pressed.
    """

    def __init__(self, *args, button: Union[int, str], **kwargs):
        """
        :param button: Button index or name.
        """
        super().__init__(*args, button=button, **kwargs)


class JoystickButtonReleasedEvent(_JoystickEvent):
    """
    Event triggered when a joystick button is released.
    """

    def __init__(self, *args, button: Union[int, str], **kwargs):
        """
        :param button: Button index or name.
        """
        super().__init__(*args, button=button, **kwargs)


class JoystickAxisEvent(_JoystickEvent):
    """
    Event triggered when an axis value of the joystick changes.
    """

    def __init__(self, *args, axis: Union[int, str], value: int, **kwargs):
        """
        :param axis: Axis index or name.
        :param value: Axis value.
        """
        super().__init__(*args, axis=axis, value=value, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.backend.weather import WeatherBackend


class WeatherOpenweathermapBackend(WeatherBackend):
    """
    Weather forecast backend that leverages the OpenWeatherMap API.

    Triggers:

        * :class:`platypush.message.event.weather.NewWeatherConditionEvent` when there is a weather condition update

    Requires:

        * The :class:`platypush.plugins.weather.openweathermap.WeatherOpenWeatherMapPlugin` plugin configured

    """

    def __init__(self, poll_seconds: int = 60, **kwargs):
        """
        :param poll_seconds: How often the backend should check for updates (default: every minute).
        """
        super().__init__(plugin_name='weather.openweathermap', poll_seconds=poll_seconds, **kwargs)


# vim:sw=4:ts=4:et:



import json
import multiprocessing
from typing import Optional

import requests
import websocket

from platypush.context import get_bus
from platypush.message.event.gotify import GotifyMessageEvent
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.gotify import GotifyMessageSchema


class GotifyPlugin(RunnablePlugin):
    """
    Gotify integration.

    `Gotify <https://gotify.net>`_ allows you process messages and notifications asynchronously
    over your own devices without relying on 3rd-party cloud services.

    Triggers:

        * :class:`platypush.message.event.gotify.GotifyMessageEvent` when a new message is received.

    """

    def __init__(self, server_url: str, app_token: str, client_token: str, **kwargs):
        """
        :param server_url: Base URL of the Gotify server (e.g. ``http://localhost``).
        :param app_token: Application token, required to send message and retrieve application info.
            You can create a new application under ``http://<server_url>/#/applications``.
        :param client_token: Client token, required to subscribe to messages.
            You can create a new client under ``http://<server_url>/#/clients``.
        """
        super().__init__(**kwargs)
        self.server_url = server_url
        self.app_token = app_token
        self.client_token = client_token
        self._ws_app: Optional[websocket.WebSocketApp] = None
        self._state_lock = multiprocessing.RLock()
        self._connected_event = multiprocessing.Event()
        self._disconnected_event = multiprocessing.Event()
        self._ws_listener: Optional[multiprocessing.Process] = None

    def _execute(self, method: str, endpoint: str, **kwargs) -> dict:
        method = method.lower()
        rs = getattr(requests, method)(
            f'{self.server_url}/{endpoint}',
            headers={
                'X-Gotify-Key': self.app_token if method == 'post' else self.client_token,
                'Content-Type': 'application/json',
                **kwargs.pop('headers', {}),
            },
            **kwargs
        )

        rs.raise_for_status()
        try:
            return rs.json()
        except Exception as e:
            self.logger.debug(e)

    def main(self):
        self._connect()
        stop_events = []

        while not any(stop_events):
            stop_events = self._should_stop.wait(timeout=1), self._disconnected_event.wait(timeout=1)

    def stop(self):
        if self._ws_app:
            self._ws_app.close()
            self._ws_app = None

        if self._ws_listener and self._ws_listener.is_alive():
            self.logger.info('Terminating websocket process')
            self._ws_listener.terminate()
            self._ws_listener.join(5)

        if self._ws_listener and self._ws_listener.is_alive():
            self.logger.warning('Terminating the websocket process failed, killing the process')
            self._ws_listener.kill()

        if self._ws_listener:
            self._ws_listener.join()
            self._ws_listener = None

        super().stop()

    def _connect(self):
        with self._state_lock:
            if self.should_stop() or self._connected_event.is_set():
                return

            ws_url = '/'.join([self.server_url.split('/')[0].replace('http', 'ws'), *self.server_url.split('/')[1:]])
            self._ws_app = websocket.WebSocketApp(
                f'{ws_url}/stream?token={self.client_token}',
                on_open=self._on_open(),
                on_message=self._on_msg(),
                on_error=self._on_error(),
                on_close=self._on_close()
            )

            def server():
                self._ws_app.run_forever()

            self._ws_listener = multiprocessing.Process(target=server)
            self._ws_listener.start()

    def _on_open(self):
        def hndl(*_):
            with self._state_lock:
                self._disconnected_event.clear()
                self._connected_event.set()
            self.logger.info('Connected to the Gotify websocket')

        return hndl

    @staticmethod
    def _on_msg():
        def hndl(*args):
            data = json.loads(args[1] if len(args) > 1 else args[0])
            get_bus().post(GotifyMessageEvent(**GotifyMessageSchema().dump(data)))

        return hndl

    def _on_error(self):
        def hndl(*args):
            error = args[1] if len(args) > 1 else args[0]
            ws = args[0] if len(args) > 1 else None
            self.logger.warning('Gotify websocket error: {}'.format(error))
            if ws:
                ws.close()

        return hndl

    def _on_close(self):
        def hndl(*_):
            with self._state_lock:
                self._disconnected_event.set()
                self._connected_event.clear()
            self.logger.warning('Gotify websocket connection closed')

        return hndl

    @action
    def send_message(self, message: str, title: Optional[str] = None, priority: int = 0, extras: Optional[dict] = None):
        """
        Send a message to the server.

        :param message: Message body (Markdown is supported).
        :param title: Message title.
        :param priority: Message priority (default: 0).
        :param extras: Extra JSON payload to be passed on the message.
        :return: .. schema:: gotify.GotifyMessageSchema
        """
        return GotifyMessageSchema().dump(
            self._execute('post', 'message', json={
                'message': message,
                'title': title,
                'priority': priority,
                'extras': extras or {},
            })
        )

    @action
    def get_messages(self, limit: int = 100, since: Optional[int] = None):
        """
        Get a list of the messages received on the server.

        :param limit: Maximum number of messages to retrieve (default: 100).
        :param since: Retrieve the message having ``since`` as minimum ID.
        :return: .. schema:: gotify.GotifyMessageSchema(many=True)
        """
        return GotifyMessageSchema().dump(
            self._execute(
                'get', 'message', params={
                    'limit': limit,
                    **({'since': since} if since else {}),
                }
            ).get('messages', []), many=True
        )

    @action
    def delete_messages(self, *ids):
        """
        Delete messages.

        :param ids: If specified, it deletes the messages matching these IDs.
            Otherwise, it deletes all the received messages.
        """
        if not ids:
            self._execute('delete', 'message')
            return

        for id in ids:
            self._execute('delete', f'message/{id}')


# vim:sw=4:ts=4:et:



from flask import Blueprint, render_template

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate

from platypush.backend.http.utils import HttpUtils

panel = Blueprint('plugin', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    panel,
]


@panel.route('/plugin/<plugin>', methods=['GET'])
@authenticate()
def plugin_route(plugin):
    """ Route to the plugin panel template """
    return render_template('index.html', plugin=plugin, utils=HttpUtils)


# vim:sw=4:ts=4:et:



import json
import logging
import os
import pathlib
import re
from base64 import b64encode
from datetime import datetime, timedelta
from random import randint
from typing import Optional, Iterable
from urllib import parse

import requests

from platypush.config import Config
from platypush.context import get_backend
from platypush.exceptions import PlatypushException
from platypush.schemas.spotify import SpotifyTrackSchema
from platypush.utils import get_ip_or_hostname, get_redis


class MissingScopesException(PlatypushException):
    """
    Exception raised in case of insufficient access scopes for an API call.
    """
    def __init__(self, scopes: Optional[Iterable[str]] = None):
        super().__init__('Missing scopes for the required API call')
        self.scopes = scopes

    def __str__(self):
        return f'{self._msg}: {self.scopes}'


class SpotifyMixin:
    """
    This mixin provides a common interface to access the Spotify API.
    """

    spotify_auth_redis_prefix = 'platypush/music/spotify/auth'
    spotify_required_scopes = [
        'user-read-playback-state',
        'user-modify-playback-state',
        'user-read-currently-playing',
        'user-read-recently-played',
        'app-remote-control',
        'streaming',
        'playlist-modify-public',
        'playlist-modify-private',
        'playlist-read-private',
        'playlist-read-collaborative',
        'user-library-modify',
        'user-library-read',
    ]

    def __init__(self, *, client_id: Optional[str] = None, client_secret: Optional[str] = None, **kwargs):
        """
        :param client_id: Spotify client ID.
        :param client_secret: Spotify client secret.
        """
        self._spotify_data_dir = os.path.join(os.path.expanduser(Config.get('workdir')), 'spotify')
        self._spotify_credentials_file = os.path.join(self._spotify_data_dir, 'credentials.json')
        self._spotify_api_token: Optional[str] = None
        self._spotify_api_token_expires_at: Optional[datetime] = None
        self._spotify_user_token: Optional[str] = None
        self._spotify_user_token_expires_at: Optional[datetime] = None
        self._spotify_user_scopes = set()
        self._spotify_refresh_token: Optional[str] = None

        if not (client_id and client_secret) and Config.get('backend.music.spotify'):
            client_id, client_secret = (
                Config.get('backend.music.spotify').get('client_id'),
                Config.get('backend.music.spotify').get('client_secret'),
            )

        if not (client_id and client_secret) and Config.get('music.spotify'):
            client_id, client_secret = (
                Config.get('music.spotify').get('client_id'),
                Config.get('music.spotify').get('client_secret'),
            )

        self._spotify_api_credentials = (client_id, client_secret) if client_id and client_secret else ()
        self._spotify_logger = logging.getLogger(__name__)
        pathlib.Path(self._spotify_data_dir).mkdir(parents=True, exist_ok=True)

    def _spotify_assert_keys(self):
        assert self._spotify_api_credentials, \
            'No Spotify API credentials provided. ' + \
            'Please register an app on https://developers.spotify.com'

    def _spotify_authorization_header(self) -> dict:
        self._spotify_assert_keys()
        return {
            'Authorization': 'Basic ' + b64encode(
                    f'{self._spotify_api_credentials[0]}:{self._spotify_api_credentials[1]}'.encode()
                ).decode()
        }

    def _spotify_load_user_credentials(self, scopes: Optional[Iterable[str]] = None):
        if not os.path.isfile(self._spotify_credentials_file):
            return

        with open(self._spotify_credentials_file, 'r') as f:
            credentials = json.load(f)

        access_token, refresh_token, expires_at, saved_scopes = (
            credentials.get('access_token'),
            credentials.get('refresh_token'),
            credentials.get('expires_at'),
            set(credentials.get('scopes', [])),
        )

        self._spotify_refresh_token = refresh_token
        self._spotify_user_scopes = self._spotify_user_scopes.union(saved_scopes)

        if not expires_at:
            self._spotify_user_token = None
            self._spotify_user_token_expires_at = None
            return

        expires_at = datetime.fromisoformat(expires_at)
        if expires_at <= datetime.now():
            self._spotify_user_token = None
            self._spotify_user_token_expires_at = None
            return

        missing_scopes = [scope for scope in (scopes or []) if scope not in saved_scopes]
        if missing_scopes:
            self._spotify_user_token = None
            self._spotify_user_token_expires_at = None
            raise MissingScopesException(scopes=missing_scopes)

        self._spotify_user_token = access_token
        self._spotify_user_token_expires_at = expires_at

    def _spotify_save_user_credentials(self, access_token: str, refresh_token: str, expires_at: datetime,
                                       scopes: Optional[Iterable[str]] = None):
        self._spotify_user_token = access_token
        self._spotify_user_token_expires_at = expires_at
        self._spotify_refresh_token = refresh_token
        self._spotify_user_scopes = self._spotify_user_scopes.union(set(scopes or []))

        with open(self._spotify_credentials_file, 'w') as f:
            os.chmod(self._spotify_credentials_file, 0o600)
            json.dump({
                'access_token': access_token,
                'refresh_token': refresh_token,
                'expires_at': datetime.isoformat(expires_at),
                'scopes': list(self._spotify_user_scopes),
            }, f)

    def spotify_api_authenticate(self):
        """
        Authenticate to the Spotify API for requests that don't require access to user data.
        """
        if not (self._spotify_api_token or self._spotify_user_token):
            self._spotify_load_user_credentials()

        self._spotify_assert_keys()
        if (self._spotify_user_token and self._spotify_user_token_expires_at > datetime.now()) or \
                (self._spotify_api_token and self._spotify_api_token_expires_at > datetime.now()):
            # Already authenticated
            return

        rs = requests.post(
            'https://accounts.spotify.com/api/token',
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
                **self._spotify_authorization_header(),
            },
            data={
                'grant_type': 'client_credentials',
            }
        )

        rs.raise_for_status()
        rs = rs.json()
        self._spotify_api_token = rs.get('access_token')
        self._spotify_api_token_expires_at = datetime.now() + timedelta(seconds=rs.get('expires_in'))

    def spotify_user_authenticate(self, scopes: Optional[Iterable[str]] = None):
        """
        Authenticate to the Spotify API for requests that require access to user data.
        """
        if self._spotify_user_token:
            return

        try:
            self._spotify_load_user_credentials(scopes=scopes or self.spotify_required_scopes)
            if self._spotify_user_token:
                return

            if self._spotify_refresh_token:
                try:
                    self._spotify_refresh_user_token()
                    return
                except Exception as e:
                    self._spotify_logger.error(f'Unable to refresh the user access token: {e}')
        except MissingScopesException as e:
            self._spotify_logger.warning(e)

        http = get_backend('http')
        assert http, 'HTTP backend not configured'
        callback_url = '{scheme}://{host}:{port}/spotify/auth_callback'.format(
            scheme="https" if http.ssl_context else "http",
            host=get_ip_or_hostname(),
            port=http.port,
        )

        state = b64encode(bytes([randint(0, 255) for _ in range(18)])).decode()
        self._spotify_logger.warning('\n\nUnauthenticated Spotify session or scopes not provided by the user. Please '
                                     'open the following URL in a browser to authenticate:\n'
                                     'https://accounts.spotify.com/authorize?client_id='
                                     f'{self._spotify_api_credentials[0]}&'
                                     f'response_type=code&redirect_uri={parse.quote(callback_url, safe="")}'
                                     f'&scope={parse.quote(" ".join(scopes))}&state={state}.\n'
                                     'Replace the host in the callback URL with the IP/hostname of this machine '
                                     f'accessible to your browser if required, and make sure to add {callback_url} '
                                     'to the list of whitelisted callbacks on your Spotify application page.\n')

        redis = get_redis()
        msg = json.loads(redis.blpop(self.get_spotify_queue_for_state(state))[1].decode())
        assert not msg.get('error'), f'Authentication error: {msg["error"]}'
        self._spotify_user_authenticate_phase_2(code=msg['code'], callback_url=callback_url, scopes=scopes)

    def _spotify_user_authenticate_phase_2(self, code: str, callback_url: str, scopes: Iterable[str]):
        rs = requests.post(
            'https://accounts.spotify.com/api/token',
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
                **self._spotify_authorization_header(),
            },
            data={
                'code': code,
                'redirect_uri': callback_url,
                'grant_type': 'authorization_code',
            }
        )

        rs.raise_for_status()
        rs = rs.json()
        self._spotify_save_user_credentials(access_token=rs.get('access_token'),
                                            refresh_token=rs.get('refresh_token'),
                                            scopes=scopes,
                                            expires_at=datetime.now() + timedelta(seconds=rs['expires_in']))

    def _spotify_refresh_user_token(self):
        self._spotify_logger.debug('Refreshing user access token')
        rs = requests.post(
            'https://accounts.spotify.com/api/token',
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
                **self._spotify_authorization_header(),
            },
            data={
                'refresh_token': self._spotify_refresh_token,
                'grant_type': 'refresh_token',
            }
        )

        rs.raise_for_status()
        rs = rs.json()
        self._spotify_save_user_credentials(access_token=rs.get('access_token'),
                                            refresh_token=rs.get('refresh_token', self._spotify_refresh_token),
                                            expires_at=datetime.now() + timedelta(seconds=rs['expires_in']))

    @classmethod
    def get_spotify_queue_for_state(cls, state: str):
        return cls.spotify_auth_redis_prefix + '/' + state

    def spotify_user_call(self, url: str, method='get', scopes: Optional[Iterable[str]] = None, **kwargs) -> dict:
        """
        Shortcut for ``spotify_api_call`` that requires all the application scopes if none are passed.
        """
        return self.spotify_api_call(url, method=method, scopes=scopes or self.spotify_required_scopes, **kwargs)

    def spotify_api_call(self, url: str, method='get', scopes: Optional[Iterable[str]] = None, **kwargs) -> dict:
        """
        Send an API request to a Spotify endpoint.

        :param url: URL to be requested.
        :param method: HTTP method (default: ``get``).
        :param scopes: List of scopes required by the call.
        :param kwargs: Extra keyword arguments to be passed to the request.
        :return: The response payload.
        """
        if scopes:
            self.spotify_user_authenticate(scopes=scopes)
        else:
            self.spotify_api_authenticate()

        method = getattr(requests, method.lower())
        rs = method(
            f'https://api.spotify.com{url}',
            headers={
                'Authorization': f'Bearer {self._spotify_user_token or self._spotify_api_token}',
                'Content-Type': 'application/json',
                'Accept': 'application/json',
            },
            **kwargs
        )

        rs.raise_for_status()
        if rs.status_code != 204 and rs.text:
            return rs.json()

    def spotify_get_track(self, track_id: str):
        """
        Get information about a Spotify track ID.
        """
        if self._spotify_api_credentials:
            info = self.spotify_api_call(f'/v1/tracks/{track_id}')
        else:
            info = json.loads(
                [
                    re.match(r'^\s*Spotify.Entity\s*=\s*(.*);\s*$', line).group(1)
                    for line in requests.get(f'https://open.spotify.com/track/{track_id}').text.split('\n')
                    if 'Spotify.Entity' in line
                ].pop()
            )

        return SpotifyTrackSchema().dump({
            **info,
            'file': info['uri'],
            'time': info['duration_ms']/1000. if info.get('duration_ms') is not None else None,
            'artist': '; '.join([
                artist['name'] for artist in info.get('artists', [])
            ]),
            'album': info.get('album', {}).get('name'),
            'name': info.get('name'),
            'title': info.get('name'),
            'date': int(info.get('album', {}).get('release_date', '').split('-')[0]),
            'track': info.get('track_number'),
            'id': info['id'],
            'x-albumuri': info.get('album', {}).get('uri'),
        })

    # noinspection PyShadowingBuiltins
    def _spotify_paginate_results(self, url: str, limit: Optional[int] = None, offset: Optional[int] = None,
                                  type: Optional[str] = None, **kwargs) -> Iterable:
        results = []

        while url and (limit is None or len(results) < limit):
            url = parse.urlparse(url)
            kwargs['params'] = {
                **kwargs.get('params', {}),
                **({'limit': min(limit, 50)} if limit is not None else {}),
                **({'offset': offset} if offset is not None else {}),
                **parse.parse_qs(url.query),
            }

            page = self.spotify_user_call(url.path, **kwargs)
            if type:
                page = page.pop(type + 's')
            results.extend(page.pop('items', []) if isinstance(page, dict) else page)
            url = page.pop('next', None) if isinstance(page, dict) else None

        return results[:limit]



from datetime import datetime
from typing import Optional

from platypush.message.event import Event


class Covid19UpdateEvent(Event):
    def __init__(self,
                 confirmed: int,
                 deaths: int,
                 recovered: int,
                 country: Optional[str] = None,
                 country_code: Optional[str] = None,
                 update_time: Optional[datetime] = None,
                 *args, **kwargs):
        super().__init__(*args,
                         confirmed=confirmed,
                         deaths=deaths,
                         recovered=recovered,
                         country=country,
                         country_code=country_code,
                         update_time=update_time,
                         **kwargs)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, List, Union

from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin


class ZwaveBasePlugin(SwitchPlugin, ABC):
    """
    Base class for Z-Wave plugins.
    """

    @abstractmethod
    @action
    def start_network(self):
        raise NotImplementedError

    @abstractmethod
    @action
    def stop_network(self):
        raise NotImplementedError

    @abstractmethod
    @action
    def status(self) -> Dict[str, Any]:
        """
        Get the status of the controller.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def add_node(self, *args, **kwargs):
        """
        Start the inclusion process to add a node to the network.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def remove_node(self):
        """
        Remove a node from the network.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def remove_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Remove a failed node from the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def replace_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Replace a failed node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def replication_send(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Send node information from the primary to the secondary controller.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def request_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Request a network update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def request_node_neighbour_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Request a neighbours list update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_nodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the nodes associated to the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_node_stats(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Get the statistics of a node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_node_name(self, new_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None):
        """
        Rename a node on the network.

        :param new_name: New name for the node.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_node_product_name(self, product_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None,
                              **kwargs):
        """
        Set the product name of a node.

        :param product_name: Product name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_node_manufacturer_name(self, manufacturer_name: str, node_id: Optional[int] = None,
                                   node_name: Optional[str] = None, **kwargs):
        """
        Set the manufacturer name of a node.

        :param manufacturer_name: Manufacturer name.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_node_location(self, location: str, node_id: Optional[int] = None, node_name: Optional[str] = None,
                          **kwargs):
        """
        Set the location of a node.

        :param location: Node location.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def cancel_command(self):
        """
        Cancel the current running command.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def kill_command(self):
        """
        Immediately terminate any running command on the controller and release the lock.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_controller_name(self, name: str, **kwargs):
        """
        Set the name of the controller on the network.

        :param name: New controller name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_capabilities(self, **kwargs) -> List[str]:
        """
        Get the capabilities of the controller.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def receive_configuration(self, **kwargs):
        """
        Receive the configuration from the primary controller on the network. Requires a primary controller active.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def transfer_primary_role(self, **kwargs):
        """
        Add a new controller to the network and make it the primary.
        The existing primary will become a secondary controller.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def heal(self, refresh_routes: bool = False, **kwargs):
        """
        Heal network by requesting nodes rediscover their neighbors.

        :param refresh_routes: Whether to perform return routes initialization (default: ``False``).
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def switch_all(self, state: bool, **kwargs):
        """
        Switch all the connected devices on/off.

        :param state: True (switch on) or False (switch off).
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def test(self, count: int = 1, **kwargs):
        """
        Send a number of test messages to every node and record results.

        :param count: The number of test messages to send.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None,
                  **kwargs) -> Dict[str, Any]:
        """
        Get a value on the network.

        :param value_id: Select by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_value(self, data, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None,
                  **kwargs):
        """
        Set a value.

        :param data: Data to set for the value.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_value_label(self, new_label: str, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None, **kwargs):
        """
        Change the label/name of a value.

        :param new_label: New value label.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_add_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                       value_label: Optional[str] = None, node_id: Optional[int] = None,
                       node_name: Optional[str] = None, **kwargs):
        """
        Add a value to a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                          value_label: Optional[str] = None, node_id: Optional[int] = None,
                          node_name: Optional[str] = None, **kwargs):
        """
        Remove a value from a node.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_heal(self, node_id: Optional[int] = None, node_name: Optional[str] = None, refresh_routes: bool = False,
                  **kwargs):
        """
        Heal network node by requesting the node to rediscover their neighbours.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param refresh_routes: Whether to perform return routes initialization. (default: ``False``).
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_update_neighbours(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Ask a node to update its neighbours table.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Update the controller with network information.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def node_refresh_info(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Fetch up-to-date information about the node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_dimmers(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the dimmers on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_node_config(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[int, Any]:
        """
        Get the configuration values of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_battery_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[int, Any]:
        """
        Get the battery levels of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_power_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[int, Any]:
        """
        Get the power levels of this node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_bulbs(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the bulbs/LEDs on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_switches(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the switches on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_sensors(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the sensors on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_doorlocks(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the doorlocks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_usercodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[int, Any]:
        """
        Get the usercodes on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_thermostats(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[int, Any]:
        """
        Get the thermostats on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_protections(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[int, Any]:
        """
        Get the protection-compatible devices on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_groups(self, **kwargs) -> Dict[int, Any]:
        """
        Get the groups on the network.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_scenes(self, **kwargs) -> Dict[str, Any]:
        """
        Get the scenes configured on the network.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def create_scene(self, label: str, **kwargs):
        """
        Create a new scene.

        :param label: Scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def remove_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs):
        """
        Remove a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def activate_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs):
        """
        Activate a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def set_scene_label(self, new_label: str, scene_id: Optional[int] = None, scene_label: Optional[str] = None,
                        **kwargs):
        """
        Rename a scene/set the scene label.

        :param new_label: New label.
        :param scene_id: Select by scene_id.
        :param scene_label: Select by current scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def scene_add_value(self, data: Optional[Any] = None,
                        value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                        value_label: Optional[str] = None, scene_id: Optional[int] = None,
                        scene_label: Optional[str] = None, node_id: Optional[int] = None,
                        node_name: Optional[str] = None, **kwargs):
        """
        Add a value to a scene.

        :param data: Data to set for the value (default: current value data).
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def scene_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                           value_label: Optional[str] = None, scene_id: Optional[int] = None,
                           scene_label: Optional[str] = None, node_id: Optional[int] = None,
                           node_name: Optional[str] = None, **kwargs):
        """
        Remove a value from a scene.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :return: value_id -> value (as a dict) mapping.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def create_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None,
                      **kwargs):
        """
        Create a handheld button on a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def delete_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None,
                      **kwargs):
        """
        Delete a button association from a device. Only intended for bridge firmware controllers.

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def add_node_to_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                          node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Add a node to a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def remove_node_from_group(self, group_index: Optional[int] = None, group_label: Optional[str] = None,
                               node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Remove a node from a group.

        :param group_index: Select group by group index.
        :param group_label: Select group by group label.
        :param node_id: Select node by node_id.
        :param node_name: Select node by node name.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def create_new_primary(self, **kwargs):
        """
        Create a new primary controller on the network when the previous primary fails.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def hard_reset(self, **kwargs):
        """
        Perform a hard reset of the controller. It erases its network configuration settings.
        The controller becomes a primary controller ready to add devices to a new network.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def soft_reset(self, **kwargs):
        """
        Perform a soft reset of the controller.
        Resets a controller without erasing its network configuration settings.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def write_config(self, **kwargs):
        """
        Store the current configuration of the network to the user directory.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def on(self, device: str, *args, **kwargs):
        """
        Turn on a switch on a device.

        :param device: ``id_on_network`` of the value to be switched on.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def off(self, device: str, *args, **kwargs):
        """
        Turn off a switch on a device.

        :param device: ``id_on_network`` of the value to be switched off.
        """
        raise NotImplementedError

    @abstractmethod
    @action
    def toggle(self, device: str, *args, **kwargs):
        """
        Toggle a switch on a device.

        :param device: ``id_on_network`` of the value to be toggled.
        """
        raise NotImplementedError


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Union

from marshmallow import fields, pre_dump
from marshmallow.schema import Schema
from marshmallow.validate import OneOf, Range

from platypush.plugins.media import PlayerState
from platypush.schemas import normalize_datetime

device_types = [
    'Unknown',
    'Computer',
    'Tablet',
    'Smartphone',
    'Speaker',
    'TV',
    'AVR',
    'STB',
    'Audio dongle',
]


class SpotifySchema(Schema):
    @staticmethod
    def _normalize_timestamp(t: Union[str, datetime]) -> datetime:
        if isinstance(t, str):
            # Replace the "Z" suffix with "+00:00"
            t = datetime.fromisoformat(t[:-1] + '+00:00')

        return t

    @pre_dump
    def _extract_url(self, data, **_):
        url = data.get('external_urls', {}).get('spotify')
        if url:
            data['url'] = url

        return data


class SpotifyDeviceSchema(SpotifySchema):
    id = fields.String(required=True, dump_only=True, metadata=dict(description='Device unique ID'))
    name = fields.String(required=True, metadata=dict(description='Device name'))
    type = fields.String(attribute='deviceType', required=True, validate=OneOf(device_types),
                         metadata=dict(description=f'Supported types: [{", ".join(device_types)}]'))
    volume = fields.Int(attribute='volume_percent', validate=Range(min=0, max=100),
                        metadata=dict(description='Player volume in percentage [0-100]'))
    is_active = fields.Boolean(required=True, dump_only=True,
                               metadata=dict(description='True if the device is currently active'))
    is_restricted = fields.Boolean(required=True, metadata=dict(description='True if the device has restricted access'))
    is_private_session = fields.Boolean(required=False,
                                        metadata=dict(description='True if the device is currently playing a private '
                                                                  'session'))


class SpotifyTrackSchema(SpotifySchema):
    id = fields.String(required=True, dump_only=True, metadata=dict(description='Spotify ID'))
    uri = fields.String(required=True, dump_only=True, metadata=dict(description='Spotify URI'))
    url = fields.String(dump_only=True, metadata=dict(description='Spotify URL'))
    file = fields.String(required=True, dump_only=True,
                         metadata=dict(description='Cross-compatibility file ID (same as uri)'))
    title = fields.String(attribute='name', required=True, metadata=dict(description='Track title'))
    artist = fields.String(metadata=dict(description='Track artist'))
    album = fields.String(metadata=dict(description='Track album'))
    image_url = fields.String(metadata=dict(description='Album image URL'))
    date = fields.Int(metadata=dict(description='Track year release date'))
    track = fields.Int(attribute='track_number', metadata=dict(description='Album track number'))
    duration = fields.Float(metadata=dict(description='Track duration in seconds'))
    popularity = fields.Int(metadata=dict(description='Popularity between 0 and 100'))
    type = fields.Constant('track', metadata=dict(description='track'))

    @pre_dump
    def normalize_fields(self, data, **_):
        album = data.pop('album', {})
        if album and isinstance(album, dict):
            data['album'] = album['name']
            data['date'] = int(album.get('release_date', '').split('-')[0])
            data['x-albumuri'] = album['uri']
            if album.get('images'):
                data['image_url'] = album['images'][0]['url']

        artists = data.pop('artists', [])
        if artists:
            data['artist'] = '; '.join([
                artist['name'] for artist in artists
            ])

        duration_ms = data.pop('duration_ms', None)
        if duration_ms:
            data['duration'] = duration_ms/1000.

        return data


class SpotifyAlbumSchema(SpotifySchema):
    id = fields.String(required=True, dump_only=True, metadata=dict(description='Spotify ID'))
    uri = fields.String(required=True, dump_only=True, metadata=dict(description='Spotify URI'))
    url = fields.String(dump_only=True, metadata=dict(description='Spotify URL'))
    name = fields.String(required=True, metadata=dict(description='Name/title'))
    artist = fields.String(metadata=dict(description='Artist'))
    image_url = fields.String(metadata=dict(description='Image URL'))
    date = fields.Int(metadata=dict(description='Release date'))
    tracks = fields.Nested(SpotifyTrackSchema, many=True, metadata=dict(description='List of tracks on the album'))
    popularity = fields.Int(metadata=dict(description='Popularity between 0 and 100'))
    type = fields.Constant('album', metadata=dict(description='album'))

    @pre_dump
    def normalize(self, data, **_):
        album = data.pop('album', data)
        tracks = album.pop('tracks', {}).pop('items', [])
        if tracks:
            album['tracks'] = tracks

        artists = album.pop('artists', [])
        if artists:
            album['artist'] = ';'.join([artist['name'] for artist in artists])

        date = album.pop('release_date', None)
        if date:
            album['date'] = date.split('-')[0]

        images = album.pop('images', [])
        if images:
            album['image_url'] = images[0]['url']

        return album


class SpotifyUserSchema(SpotifySchema):
    id = fields.String(required=True, dump_only=True)
    display_name = fields.String(required=True)
    uri = fields.String(required=True, dump_only=True)


class SpotifyPlaylistTrackSchema(SpotifyTrackSchema):
    position = fields.Int(validate=Range(min=1), metadata=dict(description='Position of the track in the playlist'))
    added_at = fields.DateTime(metadata=dict(description='When the track was added to the playlist'))
    added_by = fields.Nested(SpotifyUserSchema, metadata=dict(description='User that added the track'))
    type = fields.Constant('playlist', metadata=dict(description='playlist'))


class SpotifyStatusSchema(SpotifySchema):
    device_id = fields.String(required=True, dump_only=True, metadata=dict(description='Playing device unique ID'))
    device_name = fields.String(required=True, metadata=dict(description='Playing device name'))
    state = fields.String(required=True, validate=OneOf([s.value for s in PlayerState]),
                          metadata=dict(description=f'Supported types: [{", ".join([s.value for s in PlayerState])}]'))
    volume = fields.Int(validate=Range(min=0, max=100), required=False,
                        metadata=dict(description='Player volume in percentage [0-100]'))
    elapsed = fields.Float(required=False, metadata=dict(description='Time elapsed into the current track'))
    time = fields.Float(required=False, metadata=dict(description='Duration of the current track'))
    repeat = fields.Boolean(metadata=dict(description='True if the device is in repeat mode'))
    random = fields.Boolean(attribute='shuffle_state',
                            metadata=dict(description='True if the device is in shuffle mode'))
    track = fields.Nested(SpotifyTrackSchema, metadata=dict(description='Information about the current track'))

    @pre_dump
    def normalize_fields(self, data, **_):
        device = data.pop('device', {})
        if device:
            data['device_id'] = device['id']
            data['device_name'] = device['name']
            if device.get('volume_percent') is not None:
                data['volume'] = device['volume_percent']

        elapsed = data.pop('progress_ms', None)
        if elapsed is not None:
            data['elapsed'] = int(elapsed)/1000.

        track = data.pop('item', {})
        if track:
            data['track'] = track
            duration = track.get('duration_ms')
            if duration is not None:
                data['time'] = int(duration)/1000.

        is_playing = data.pop('is_playing', None)
        if is_playing is True:
            data['state'] = PlayerState.PLAY.value
        elif is_playing is False:
            data['state'] = PlayerState.PAUSE.value

        repeat = data.pop('repeat_state', None)
        data['repeat'] = False if (not repeat or repeat == 'off') else True
        return data


class SpotifyHistoryItemSchema(SpotifyTrackSchema):
    played_at = fields.DateTime(metadata=dict(description='Item play datetime'))

    @pre_dump
    def _normalize_timestamps(self, data, **_):
        played_at = data.pop('played_at', None)
        if played_at:
            data['played_at'] = self._normalize_timestamp(played_at)

        return data


class SpotifyPlaylistSchema(SpotifySchema):
    id = fields.String(required=True, dump_only=True)
    uri = fields.String(required=True, dump_only=True, metadata=dict(
        description='Playlist unique Spotify URI'
    ))
    url = fields.String(dump_only=True, metadata=dict(description='Spotify URL'))
    name = fields.String(required=True)
    description = fields.String()
    owner = fields.Nested(SpotifyUserSchema, metadata=dict(
        description='Playlist owner data'
    ))
    collaborative = fields.Boolean()
    public = fields.Boolean()
    snapshot_id = fields.String(dump_only=True, metadata=dict(
        description='Playlist snapshot ID - it changes when the playlist is modified'
    ))
    tracks = fields.Nested(SpotifyPlaylistTrackSchema, many=True, metadata=dict(
        description='List of tracks in the playlist'
    ))

    @pre_dump
    def _normalize_tracks(self, data, **_):
        if 'tracks' in data:
            if not isinstance(data['tracks'], list):
                data.pop('tracks')
            else:
                data['tracks'] = [
                    {
                        **track['track'],
                        'added_at': normalize_datetime(track.get('added_at')),
                        'added_by': track.get('added_by'),
                    }
                    if isinstance(track.get('track'), dict) else track
                    for track in data['tracks']
                ]

        return data


class SpotifyEpisodeSchema(SpotifyTrackSchema):
    description = fields.String(metadata=dict(description='Episode description'))
    show = fields.String(metadata=dict(description='Episode show name'))
    type = fields.Constant('episode', metadata=dict(description='episode'))

    @pre_dump
    def normalize_fields(self, data, **_):
        data = data.pop('episode', data)

        # Cross-compatibility with SpotifyTrackSchema
        show = data.pop('show', {})
        data['artist'] = data['album'] = data['show'] = show.get('name')
        data['x-albumuri'] = show['uri']
        images = data.pop('images', show.pop('images', []))
        if images:
            data['image_url'] = images[0]['url']

        return data


class SpotifyShowSchema(SpotifyAlbumSchema):
    description = fields.String(metadata=dict(description='Show description'))
    publisher = fields.String(metadata=dict(description='Show publisher name'))
    type = fields.Constant('show', metadata=dict(description='show'))

    @pre_dump
    def normalize_fields(self, data, **_):
        data = data.pop('show', data)

        # Cross-compatibility with SpotifyAlbumSchema
        data['artist'] = data.get('publisher', data.get('name'))
        images = data.pop('images', [])
        if images:
            data['image_url'] = images[0]['url']

        return data


class SpotifyArtistSchema(SpotifySchema):
    id = fields.String(metadata=dict(description='Spotify ID'))
    uri = fields.String(metadata=dict(description='Spotify URI'))
    url = fields.String(dump_only=True, metadata=dict(description='Spotify URL'))
    name = fields.String(metadata=dict(description='Artist name'))
    genres = fields.List(fields.String, metadata=dict(description='Artist genres'))
    popularity = fields.Int(metadata=dict(description='Popularity between 0 and 100'))
    image_url = fields.String(metadata=dict(description='Image URL'))
    type = fields.Constant('artist', metadata=dict(description='artist'))

    @pre_dump
    def normalize_fields(self, data, **_):
        data = data.pop('artist', data)
        images = data.pop('images', [])
        if images:
            data['image_url'] = images[0]['url']

        return data



from multiprocessing import Process, Queue, Event

from platypush.plugins.camera.model.writer import VideoWriter
from platypush.plugins.camera.model.writer.preview import PreviewWriter


class WxPreviewWriter(PreviewWriter, Process):
    """
    General class for managing previews from camera devices or sources of images.
    """

    def __init__(self, camera, plugin, *args, **kwargs):
        Process.__init__(self, *args, **kwargs)
        VideoWriter.__init__(self, camera=camera, plugin=plugin)
        self.app = None
        self.bitmap_queue = Queue()
        self.stopped_event = Event()

    def run(self) -> None:
        import wx
        from platypush.plugins.camera.model.writer.preview.wx.ui import Frame

        self.app = wx.App()
        frame = Frame(self)
        frame.Center()
        frame.Show()
        self.app.MainLoop()

    def close(self):
        if not self.app:
            return

        self.app.ExitMainLoop()
        self.app = None
        self.camera.preview = None
        self.bitmap_queue.close()
        self.bitmap_queue = None
        self.stopped_event.set()

    def write(self, image):
        if self.stopped_event.is_set():
            return

        try:
            self.bitmap_queue.put(image)
        except Exception as e:
            self.logger.warning('Could not add an image to the preview queue: {}'.format(str(e)))


# vim:sw=4:ts=4:et:



from functools import wraps
from logging import getLogger

from platypush.common import exec_wrapper

logger = getLogger(__name__)


def cron(cron_expression: str):
    def wrapper(f):
        f.cron = True
        f.cron_expression = cron_expression

        @wraps(f)
        def wrapped(*args, **kwargs):
            return exec_wrapper(f, *args, **kwargs)

        return wrapped

    return wrapper


# vim:sw=4:ts=4:et:



import asyncio
import json
import os
from typing import Optional, Collection, Mapping

import requests
import websockets
import websockets.exceptions

from platypush.context import get_bus
from platypush.message.event.ntfy import NotificationEvent
from platypush.plugins import AsyncRunnablePlugin, action


class NtfyPlugin(AsyncRunnablePlugin):
    """
    Ntfy integration.

    `ntfy <https://ntfy.sh/>`_ allows you to process asynchronous notification
    across multiple devices and it's compatible with the
    `UnifiedPush <https://unifiedpush.org/>` specification.

    Triggers:

        * :class:`platypush.message.event.ntfy.NotificationEvent` when a new notification is received.

    """

    def __init__(
        self,
        server_url: str = 'https://ntfy.sh',
        subscriptions: Optional[Collection[str]] = None,
        **kwargs,
    ):
        """
        :param server_url: Default ntfy instance base URL (default: ``https://ntfy.sh``).
        :param subscriptions: List of topics the plugin should subscribe to
            (default: none).
        """
        super().__init__(**kwargs)
        self._server_url = server_url
        self._ws_url = '/'.join(
            [
                self._server_url.split('/')[0].replace('http', 'ws'),
                *self._server_url.split('/')[1:],
            ]
        )

        self._subscriptions = subscriptions or []

    async def _get_ws_handler(self, url):
        reconnect_wait_secs = 1
        reconnect_wait_secs_max = 60

        while not self.should_stop():
            self.logger.debug(f'Connecting to {url}')

            try:
                async with websockets.connect(url) as ws:  # type: ignore
                    reconnect_wait_secs = 1
                    self.logger.info(f'Connected to {url}')
                    async for msg in ws:
                        try:
                            msg = json.loads(msg)
                        except json.JSONDecodeError as e:
                            self.logger.warning(
                                'Received invalid JSON message from the server: %s\n%s',
                                e,
                                msg,
                            )
                            continue

                        self.logger.debug('Received message on ntfy: %s', msg)
                        if msg.get('event') == 'message':
                            get_bus().post(
                                NotificationEvent(
                                    id=msg['id'],
                                    time=msg['time'],
                                    topic=msg['topic'],
                                    message=msg.get('message'),
                                    title=msg.get('title'),
                                    tags=msg.get('tags'),
                                    url=msg.get('click'),
                                    actions=msg.get('actions'),
                                    attachment=msg.get('attachment'),
                                )
                            )
            except websockets.exceptions.WebSocketException as e:
                self.logger.error('Websocket error: %s', e)
                await asyncio.sleep(reconnect_wait_secs)
                reconnect_wait_secs = min(
                    reconnect_wait_secs * 2, reconnect_wait_secs_max
                )

    async def listen(self):
        return await asyncio.wait(
            [
                self._get_ws_handler(f'{self._ws_url}/{sub}/ws')
                for sub in set(self._subscriptions)
            ]
        )

    @property
    def _should_start_runner(self):
        return bool(self._subscriptions)

    @action
    def send_message(
        self,
        topic: str,
        message: str = '',
        server_url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        attachment: Optional[str] = None,
        filename: Optional[str] = None,
        actions: Optional[Collection[Mapping[str, str]]] = None,
        email: Optional[str] = None,
        priority: Optional[str] = None,
        tags: Optional[Collection[str]] = None,
        schedule: Optional[str] = None,
    ):
        """
        Send a message/notification to a topic.

        :param topic: Topic where the message will be delivered.
        :param message: Text of the message to be sent.
        :param server_url: Override the default server URL.
        :param username: Set if publishing to the topic requires authentication
        :param password: Set if publishing to the topic requires authentication
        :param title: Custom notification title.
        :param url: URL that should be opened when the user clicks the
            notification. It can be an ``http(s)://`` URL, a ``mailto:`, a
            ``geo:``, a link to another ntfy topic (e.g. ``ntfy://mytopic``) or
            a Twitter link (e.g. ``twitter://user?screen_name=myname``).
        :param attachment: Attach a file or URL to the notification. It can
            either be an HTTP URL or a path to a local file.
        :param filename: If ``attachment`` is specified, you can override the
            output filename (default: same filename as the URL/path base name).
        :param actions: List of objects describing possible action buttons
            available for the notification. Supported types:

                - ``view``: Open a URL or an app when the action button is
                  clicked
                - ``http``: Send an HTTP request upon action selection.
                - ``broadcast``: Send an `Android broadcast <https://developer.android.com/guide/components/broadcasts>`
                      intent upon action selection (only available on Android).

            Example:

            .. code-block:: json

                [
                    {
                        "action": "view",
                        "label": "Open portal",
                        "url": "https://home.nest.com/",
                        "clear": true
                    },
                    {
                        "action": "http",
                        "label": "Turn down",
                        "url": "https://api.nest.com/",
                        "method": "PUT",
                        "headers": {
                            "Authorization": "Bearer abcdef..."
                        },
                        "body": "{\\"temperature\\": 65}"
                    },
                    {
                        "action": "broadcast",
                        "label": "Take picture",
                        "intent": "com.myapp.TAKE_PICTURE_INTENT",
                        "extras": {
                            "camera": "front"
                        }
                    }
                ]

        :param email: Forward the notification as an email to the specified
            address.
        :param priority: Custom notification priority. Supported values:
            ``[max, high, default, low, min]``.
        :param tags: Optional list of tags associated with the notification.
            Tag names that match emoji short codes will be rendered as emojis
            in the notification - see `here <https://ntfy.sh/docs/emojis/>` for
            a list of supported emojis.
        :param schedule: Schedule the message to be delivered at a specific
            time (for example, for reminders). Supported formats:

                - UNIX timestamps
                - Duration (e.g. ``30m``, ``3h``, ``2 days``)
                - Natural language strings (e.g. ``Tuesday, 7am`` or
                  ``tomorrow, 3pm``)

        """
        method = requests.post
        click_url = url
        url = server_url or self._server_url
        args = {}
        if username and password:
            args['auth'] = (username, password)

        if attachment and not (
            attachment.startswith('http://') or attachment.startswith('https://')
        ):
            url = f'{url}/{topic}'
            attachment = os.path.expanduser(attachment)
            filename = filename or os.path.basename(attachment)
            args['headers'] = {
                'Filename': filename,
                **({'X-Title': title} if title else {}),
                **({'X-Click': click_url} if click_url else {}),
                **({'X-Email': email} if email else {}),
                **({'X-Priority': priority} if priority else {}),
                **({'X-Tags': ','.join(tags)} if tags else {}),
                **({'X-Delay': schedule} if schedule else {}),
            }

            with open(attachment, 'rb') as f:
                args['data'] = f.read()
            method = requests.put
        else:
            args['json'] = {
                'topic': topic,
                'message': message,
                **({'title': title} if title else {}),
                **({'click': click_url} if click_url else {}),
                **({'email': email} if email else {}),
                **({'priority': priority} if priority else {}),
                **({'tags': tags} if tags else {}),
                **({'delay': schedule} if schedule else {}),
                **({'actions': actions} if actions else {}),
                **(
                    {
                        'attach': attachment,
                        'filename': (
                            filename
                            if filename
                            else attachment.split('/')[-1].split('?')[0]
                        ),
                    }
                    if attachment
                    else {}
                ),
            }

        rs = method(url, **args)
        assert rs.ok, 'Could not send message to {}: {}'.format(
            topic, rs.json().get('error', f'HTTP error: {rs.status_code}')
        )

        return rs.json()


# vim:sw=4:ts=4:et:



from flask import Blueprint, render_template

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, get_websocket_port
from platypush.backend.http.utils import HttpUtils

dashboard = Blueprint('dashboard', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    dashboard,
]


# noinspection PyUnusedLocal
@dashboard.route('/dashboard/<name>', methods=['GET'])
@authenticate()
def render_dashboard(name):
    """ Route for the dashboard """
    return render_template('index.html',
                           utils=HttpUtils,
                           websocket_port=get_websocket_port())


# vim:sw=4:ts=4:et:



import base64
import io
import os
import threading
import time
from typing import Optional, List

from platypush import Config
from platypush.context import get_bus
from platypush.message.event.qrcode import QrcodeScannedEvent
from platypush.message.response.qrcode import QrcodeGeneratedResponse, QrcodeDecodedResponse, ResultModel
from platypush.plugins import Plugin, action
from platypush.plugins.camera import CameraPlugin
from platypush.utils import get_plugin_class_by_name


class QrcodePlugin(Plugin):
    """
    Plugin to generate and scan QR and bar codes.

    Requires:

        * **numpy** (``pip install numpy``).
        * **qrcode** (``pip install 'qrcode[pil]'``) for QR generation.
        * **pyzbar** (``pip install pyzbar``) for decoding code from images.
        * **Pillow** (``pip install Pillow``) for image management.

    """

    def __init__(self, camera_plugin: Optional[str] = None, **kwargs):
        """
        :param camera_plugin: Name of the plugin that will be used as a camera to capture images (e.g.
            ``camera.cv`` or ``camera.pi``).
        """
        super().__init__(**kwargs)
        self.camera_plugin = camera_plugin
        self._capturing = threading.Event()

    def _get_camera(self, camera_plugin: Optional[str] = None, **config) -> CameraPlugin:
        camera_plugin = camera_plugin or self.camera_plugin
        if not config:
            config = Config.get(camera_plugin) or {}
        config['stream_raw_frames'] = True

        cls = get_plugin_class_by_name(camera_plugin)
        assert cls and issubclass(cls, CameraPlugin), '{} is not a valid camera plugin'.format(camera_plugin)
        return cls(**config)

    # noinspection PyShadowingBuiltins
    @action
    def generate(self, content: str, output_file: Optional[str] = None, show: bool = False,
                 format: str = 'png') -> QrcodeGeneratedResponse:
        """
        Generate a QR code.
        If you configured the :class:`platypush.backend.http.HttpBackend` then you can also generate
        codes directly from the browser through ``http://<host>:<port>/qrcode?content=...``.

        :param content: Text, URL or content of the QR code.
        :param output_file: If set then the QR code will be exported in the specified image file.
            Otherwise, a base64-encoded representation of its binary content will be returned in
            the response as ``data``.
        :param show: If True, and if the device where the application runs has an active display,
            then the generated QR code will be shown on display.
        :param format: Output image format (default: ``png``).
        :return: :class:`platypush.message.response.qrcode.QrcodeGeneratedResponse`.
        """
        import qrcode
        qr = qrcode.make(content)
        img = qr.get_image()
        ret = {
            'content': content,
            'format': format,
        }

        if show:
            img.show()
        if output_file:
            output_file = os.path.abspath(os.path.expanduser(output_file))
            img.save(output_file, format=format)
            ret['image_file'] = output_file
        else:
            f = io.BytesIO()
            img.save(f, format=format)
            ret['data'] = base64.encodebytes(f.getvalue()).decode()

        return QrcodeGeneratedResponse(**ret)

    @action
    def decode(self, image_file: str) -> QrcodeDecodedResponse:
        """
        Decode a QR code from an image file.

        :param image_file: Path of the image file.
        """
        from pyzbar import pyzbar
        from PIL import Image

        image_file = os.path.abspath(os.path.expanduser(image_file))
        img = Image.open(image_file)
        results = pyzbar.decode(img)
        return QrcodeDecodedResponse(results)

    @staticmethod
    def _convert_frame(frame):
        import numpy as np
        from PIL import Image

        assert isinstance(frame, np.ndarray), \
            'Image conversion only works with numpy arrays for now (got {})'.format(type(frame))
        mode = 'RGB'
        if len(frame.shape) > 2 and frame.shape[2] == 4:
            mode = 'RGBA'

        return Image.frombuffer(mode, (frame.shape[1], frame.shape[0]), frame, 'raw', mode, 0, 1)

    @action
    def start_scanning(self, camera_plugin: Optional[str] = None, duration: Optional[float] = None,
                       n_codes: Optional[int] = None) -> Optional[List[ResultModel]]:
        """
        Decode QR-codes and bar codes using a camera.

        Triggers:

            - :class:`platypush.message.event.qrcode.QrcodeScannedEvent` when a code is successfully scanned.

        :param camera_plugin: Camera plugin (overrides default ``camera_plugin``).
        :param duration: How long the capturing phase should run (default: until ``stop_scanning`` or app termination).
        :param n_codes: Stop after decoding this number of codes (default: None).
        :return: When ``duration`` or ``n_codes`` are specified or ``stop_scanning`` is called, it will return a list of
            :class:`platypush.message.response.qrcode.ResultModel` instances with the scanned results,
        """
        from pyzbar import pyzbar
        assert not self._capturing.is_set(), 'A capturing process is already running'

        camera = self._get_camera(camera_plugin)
        codes = []
        last_results = {}
        last_results_timeout = 10.0
        last_results_time = 0
        self._capturing.set()

        try:
            with camera:
                start_time = time.time()

                while self._capturing.is_set() \
                        and (not duration or time.time() < start_time + duration) \
                        and (not n_codes or len(codes) < n_codes):
                    output = camera.get_stream()
                    with output.ready:
                        output.ready.wait()
                        img = self._convert_frame(output.raw_frame)
                        results = pyzbar.decode(img)
                        if results:
                            results = [
                                result for result in QrcodeDecodedResponse(results).output['results']
                                if result['data'] not in last_results
                                   or time.time() >= last_results_time + last_results_timeout
                            ]

                        if results:
                            codes.extend(results)
                            get_bus().post(QrcodeScannedEvent(results=results))
                            last_results = {result['data']: result for result in results}
                            last_results_time = time.time()
        finally:
            self._capturing.clear()

        return codes

    @action
    def stop_scanning(self):
        self._capturing.clear()

# vim:sw=4:ts=4:et:



import queue
import socket
import time
from typing import List, Dict, Any, Optional, Union

from zeroconf import Zeroconf, ServiceInfo, ServiceBrowser, ServiceListener, ZeroconfServiceTypes

from platypush.context import get_bus
from platypush.message.event.zeroconf import ZeroconfServiceAddedEvent, ZeroconfServiceRemovedEvent, \
    ZeroconfServiceUpdatedEvent
from platypush.plugins import Plugin, action


class ZeroconfListener(ServiceListener):
    def __init__(self, evt_queue: queue.Queue):
        super().__init__()
        self.evt_queue = evt_queue

    @classmethod
    def get_service_info(cls, zc: Zeroconf, type_: str, name: str) -> dict:
        info = zc.get_service_info(type_, name)
        if not info:
            return {}

        return cls.parse_service_info(info)

    @staticmethod
    def parse_service_info(info: ServiceInfo) -> dict:
        return {
            'addresses': [socket.inet_ntoa(addr) for addr in info.addresses if info.addresses],
            'port': info.port,
            'host_ttl': info.host_ttl,
            'other_ttl': info.other_ttl,
            'priority': info.priority,
            'properties': {k.decode() if isinstance(k, bytes) else k: v.decode() if isinstance(v, bytes) else v
                           for k, v in info.properties.items()},
            'server': info.server,
            'weight': info.weight,
        }

    def add_service(self, zc: Zeroconf, type_: str, name: str):
        info = self.get_service_info(zc, type_, name)
        self.evt_queue.put(ZeroconfServiceAddedEvent(service_type=type_, service_name=name, service_info=info))

    def remove_service(self, zc: Zeroconf, type_: str, name: str):
        info = self.get_service_info(zc, type_, name)
        self.evt_queue.put(ZeroconfServiceRemovedEvent(service_type=type_, service_name=name, service_info=info))

    def update_service(self, zc: Zeroconf, type_: str, name: str):
        info = self.get_service_info(zc, type_, name)
        self.evt_queue.put(ZeroconfServiceUpdatedEvent(service_type=type_, service_name=name, service_info=info))


class ZeroconfPlugin(Plugin):
    """
    Plugin for Zeroconf services discovery.

    Triggers:

        * :class:`platypush.message.event.zeroconf.ZeroconfServiceAddedEvent` when a new service is discovered.
        * :class:`platypush.message.event.zeroconf.ZeroconfServiceUpdatedEvent` when a service is updated.
        * :class:`platypush.message.event.zeroconf.ZeroconfServiceRemovedEvent` when a service is removed.

    Requires:

        * **zeroconf** (``pip install zeroconf``)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._discovery_in_progress = False

    @action
    def get_services(self, timeout: int = 5) -> List[str]:
        """
        Get the full list of services found on the network.

        :param timeout: Discovery timeout in seconds (default: 5).
        :return: List of the services as strings.
        """
        return list(ZeroconfServiceTypes.find(timeout=timeout))

    @action
    def discover_service(self, service: Union[str, list], timeout: Optional[int] = 5) -> Dict[str, Any]:
        """
        Find all the services matching the specified type.

        :param service: Service type (e.g. ``_http._tcp.local.``) or list of service types.
        :param timeout: Browser timeout in seconds (default: 5). Specify None for no timeout - in such case the
            discovery will loop forever and generate events upon service changes.
        :return: A ``service_type -> [service_names]`` mapping. Example:

          .. code-block:: json

            {
                "host1._platypush-http._tcp.local.": {
                    "type": "_platypush-http._tcp.local.",
                    "name": "host1._platypush-http._tcp.local.",
                    "info": {
                        "addresses": ["192.168.1.11"],
                        "port": 8008,
                        "host_ttl": 120,
                        "other_ttl": 4500,
                        "priority": 0,
                        "properties": {
                            "name": "Platypush",
                            "vendor": "Platypush",
                            "version": "0.13.2"
                        },
                        "server": "host1._platypush-http._tcp.local.",
                        "weight": 0
                    }
                }
            }

        """
        assert not self._discovery_in_progress, 'A discovery process is already running'
        self._discovery_in_progress = True

        evt_queue = queue.Queue()
        zc = Zeroconf()
        listener = ZeroconfListener(evt_queue=evt_queue)
        discovery_start = time.time()
        services = {}
        browser = None

        try:
            browser = ServiceBrowser(zc, service, listener)

            while timeout and time.time() - discovery_start < timeout:
                to = discovery_start + timeout - time.time() if timeout else None
                try:
                    evt = evt_queue.get(block=True, timeout=to)
                    if isinstance(evt, ZeroconfServiceAddedEvent) or isinstance(evt, ZeroconfServiceUpdatedEvent):
                        services[evt.service_name] = {
                            'type': evt.service_type,
                            'name': evt.service_name,
                            'info': evt.service_info,
                        }
                    elif isinstance(evt, ZeroconfServiceRemovedEvent):
                        if evt.service_name in services:
                            del services[evt.service_name]

                    get_bus().post(evt)
                except queue.Empty:
                    if not services:
                        self.logger.warning('No such service discovered: {}'.format(service))
        finally:
            if browser:
                browser.cancel()
            zc.close()
            self._discovery_in_progress = False

        return services


# vim:sw=4:ts=4:et:



import asyncio
import logging
import os
import pathlib
import re

from dataclasses import dataclass
from typing import Collection, Coroutine, Sequence
from urllib.parse import urlparse

from nio import (
    Api,
    ErrorResponse,
    MatrixRoom,
    RoomMessage,
)

from nio.api import MessageDirection, RoomVisibility

from nio.crypto.device import OlmDevice
from nio.exceptions import OlmUnverifiedDeviceError

from platypush.config import Config

from platypush.plugins import AsyncRunnablePlugin, action
from platypush.schemas.matrix import (
    MatrixDeviceSchema,
    MatrixDownloadedFileSchema,
    MatrixEventIdSchema,
    MatrixMemberSchema,
    MatrixMessagesResponseSchema,
    MatrixMyDeviceSchema,
    MatrixProfileSchema,
    MatrixRoomIdSchema,
    MatrixRoomSchema,
)

from platypush.utils import get_mime_type

from .client import MatrixClient

logger = logging.getLogger(__name__)


@dataclass
class Credentials:
    server_url: str
    user_id: str
    access_token: str
    device_id: str | None

    def to_dict(self) -> dict:
        return {
            'server_url': self.server_url,
            'user_id': self.user_id,
            'access_token': self.access_token,
            'device_id': self.device_id,
        }


class MatrixPlugin(AsyncRunnablePlugin):
    """
    Matrix chat integration.

    Requires:

        * **matrix-nio** (``pip install 'matrix-nio[e2e]'``)
        * **libolm** (on Debian ```apt-get install libolm-devel``, on Arch
            ``pacman -S libolm``)
        * **async_lru** (``pip install async_lru``)

    Note that ``libolm`` and the ``[e2e]`` module are only required if you want
    E2E encryption support.

    Unless you configure the extension to use the token of an existing trusted
    device, it is recommended that you mark the virtual device used by this
    integration as trusted through a device that is already trusted. You may
    encounter errors when sending or receiving messages on encrypted rooms if
    your user has some untrusted devices. The easiest way to mark the device as
    trusted is the following:

        - Configure the integration with your credentials and start Platypush.
        - Use the same credentials to log in through a Matrix app or web client
          (Element, Hydrogen, etc.) that has already been trusted.
        - You should see a notification that prompts you to review the
          untrusted devices logged in to your account. Dismiss it for now -
          that verification path is currently broken on the underlying library
          used by this integration.
        - Instead, select a room that you have already joined, select the list
          of users in the room and select yourself.
        - In the _Security_ section, you should see that at least one device is
          marked as unverified, and you can start the verification process by
          clicking on it.
        - Select "*Verify through emoji*". A list of emojis should be prompted.
          Optionally, verify the logs of the application to check that you see
          the same list. Then confirm that you see the same emojis, and your
          device will be automatically marked as trusted.

    All the URLs returned by actions and events on this plugin are in the
    ``mxc://<server>/<media_id>`` format. You can either convert them to HTTP
    through the :meth:`.mxc_to_http` method, or download them through the
    :meth:`.download` method.

    Triggers:

        * :class:`platypush.message.event.matrix.MatrixMessageEvent`: when a
            message is received.
        * :class:`platypush.message.event.matrix.MatrixMessageImageEvent`: when a
            message containing an image is received.
        * :class:`platypush.message.event.matrix.MatrixMessageAudioEvent`: when a
            message containing an audio file is received.
        * :class:`platypush.message.event.matrix.MatrixMessageVideoEvent`: when a
            message containing a video file is received.
        * :class:`platypush.message.event.matrix.MatrixMessageFileEvent`: when a
            message containing a generic file is received.
        * :class:`platypush.message.event.matrix.MatrixSyncEvent`: when the
            startup synchronization has been completed and the plugin is ready to
            use.
        * :class:`platypush.message.event.matrix.MatrixRoomCreatedEvent`: when
            a room is created.
        * :class:`platypush.message.event.matrix.MatrixRoomJoinEvent`: when a
            user joins a room.
        * :class:`platypush.message.event.matrix.MatrixRoomLeaveEvent`: when a
            user leaves a room.
        * :class:`platypush.message.event.matrix.MatrixRoomInviteEvent`: when
            the user is invited to a room.
        * :class:`platypush.message.event.matrix.MatrixRoomTopicChangedEvent`:
            when the topic/title of a room changes.
        * :class:`platypush.message.event.matrix.MatrixCallInviteEvent`: when
            the user is invited to a call.
        * :class:`platypush.message.event.matrix.MatrixCallAnswerEvent`: when a
            called user wishes to pick the call.
        * :class:`platypush.message.event.matrix.MatrixCallHangupEvent`: when a
            called user exits the call.
        * :class:`platypush.message.event.matrix.MatrixEncryptedMessageEvent`:
            when a message is received but the client doesn't have the E2E keys
            to decrypt it, or encryption has not been enabled.
        * :class:`platypush.message.event.matrix.MatrixRoomTypingStartEvent`:
            when a user in a room starts typing.
        * :class:`platypush.message.event.matrix.MatrixRoomTypingStopEvent`:
            when a user in a room stops typing.
        * :class:`platypush.message.event.matrix.MatrixRoomSeenReceiptEvent`:
            when the last message seen by a user in a room is updated.
        * :class:`platypush.message.event.matrix.MatrixUserPresenceEvent`:
            when a user comes online or goes offline.

    """

    def __init__(
        self,
        server_url: str = 'https://matrix-client.matrix.org',
        user_id: str | None = None,
        password: str | None = None,
        access_token: str | None = None,
        device_name: str | None = 'platypush',
        device_id: str | None = None,
        download_path: str | None = None,
        autojoin_on_invite: bool = True,
        autotrust_devices: bool = False,
        autotrust_devices_whitelist: Collection[str] | None = None,
        autotrust_users_whitelist: Collection[str] | None = None,
        autotrust_rooms_whitelist: Collection[str] | None = None,
        **kwargs,
    ):
        """
        Authentication requires user_id/password on the first login.
        Afterwards, session credentials are stored under
        ``<$PLATYPUSH_WORKDIR>/matrix/credentials.json`` (default:
        ``~/.local/share/platypush/matrix/credentials.json``), and you can
        remove the cleartext credentials from your configuration file.

        Otherwise, if you already have an ``access_token``, you can set the
        associated field instead of using ``password``. This may be required if
        the user has 2FA enabled.

        :param server_url: Default Matrix instance base URL (default:
            ``https://matrix-client.matrix.org``).
        :param user_id: user_id, in the format ``@user:example.org``, or just
            the username if the account is hosted on the same server configured in
            the ``server_url``.
        :param password: User password.
        :param access_token: User access token.
        :param device_name: The name of this device/connection (default: ``platypush``).
        :param device_id: Use an existing ``device_id`` for the sessions.
        :param download_path: The folder where downloaded media will be saved
            (default: ``~/Downloads``).
        :param autojoin_on_invite: Whether the account should automatically
            join rooms upon invite. If false, then you may want to implement your
            own logic in an event hook when a
            :class:`platypush.message.event.matrix.MatrixRoomInviteEvent` event is
            received, and call the :meth:`.join` method if required.
        :param autotrust_devices: If set to True, the plugin will automatically
            trust the devices on encrypted rooms. Set this property to True
            only if you only plan to use a bot on trusted rooms. Note that if
            no automatic trust mechanism is set you may need to explicitly
            create your logic for trusting users - either with a hook when
            :class:`platypush.message.event.matrix.MatrixSyncEvent` is
            received, or when a room is joined, or before sending a message.
        :param autotrust_devices_whitelist: Automatically trust devices with IDs
            IDs provided in this list.
        :param autotrust_users_whitelist: Automatically trust devices from the
            user IDs provided in this list.
        :param autotrust_rooms_whitelist: Automatically trust devices on the
            room IDs provided in this list.
        """
        super().__init__(**kwargs)
        if not (server_url.startswith('http://') or server_url.startswith('https://')):
            server_url = f'https://{server_url}'
        self._server_url = server_url
        server_name = self._server_url.split('/')[2].split(':')[0]

        if user_id and not re.match(user_id, '^@[a-zA-Z0-9.-_]+:.+'):
            user_id = f'@{user_id}:{server_name}'

        self._user_id = user_id
        self._password = password
        self._access_token = access_token
        self._device_name = device_name
        self._device_id = device_id
        self._download_path = download_path or os.path.join(
            os.path.expanduser('~'), 'Downloads'
        )

        self._autojoin_on_invite = autojoin_on_invite
        self._autotrust_devices = autotrust_devices
        self._autotrust_devices_whitelist = set(autotrust_devices_whitelist or [])
        self._autotrust_users_whitelist = set(autotrust_users_whitelist or [])
        self._autotrust_rooms_whitelist = set(autotrust_rooms_whitelist or [])
        self._workdir = os.path.join(Config.get('workdir'), 'matrix')  # type: ignore
        self._credentials_file = os.path.join(self._workdir, 'credentials.json')
        self._processed_responses = {}
        self._client = self._get_client()
        pathlib.Path(self._workdir).mkdir(parents=True, exist_ok=True)

    def _get_client(self) -> MatrixClient:
        return MatrixClient(
            homeserver=self._server_url,
            user=self._user_id,
            credentials_file=self._credentials_file,
            autojoin_on_invite=self._autojoin_on_invite,
            autotrust_devices=self._autotrust_devices,
            autotrust_devices_whitelist=self._autotrust_devices_whitelist,
            autotrust_rooms_whitelist=self._autotrust_rooms_whitelist,
            autotrust_users_whitelist=self._autotrust_users_whitelist,
            device_id=self._device_id,
        )

    @property
    def client(self) -> MatrixClient:
        if not self._client:
            self._client = self._get_client()
        return self._client

    async def _login(self) -> MatrixClient:
        await self.client.login(
            password=self._password,
            device_name=self._device_name,
            token=self._access_token,
        )

        return self.client

    async def listen(self):
        while not self.should_stop():
            await self._login()

            try:
                await self.client.sync_forever(timeout=30000, full_state=True)
            except KeyboardInterrupt:
                pass
            except Exception as e:
                self.logger.exception(e)
                self.logger.info('Waiting 10 seconds before reconnecting')
                await asyncio.sleep(10)
            finally:
                try:
                    await self.client.close()
                finally:
                    self._client = None

    def _loop_execute(self, coro: Coroutine):
        assert self._loop, 'The loop is not running'

        try:
            ret = asyncio.run_coroutine_threadsafe(coro, self._loop).result()
        except OlmUnverifiedDeviceError as e:
            raise AssertionError(str(e))

        assert not isinstance(ret, ErrorResponse), ret.message
        if hasattr(ret, 'transport_response'):
            response = ret.transport_response
            assert response.ok, f'{coro} failed with status {response.status}'

        return ret

    def _process_local_attachment(self, attachment: str, room_id: str) -> dict:
        attachment = os.path.expanduser(attachment)
        assert os.path.isfile(attachment), f'{attachment} is not a valid file'

        filename = os.path.basename(attachment)
        mime_type = get_mime_type(attachment) or 'application/octet-stream'
        message_type = mime_type.split('/')[0]
        if message_type not in {'audio', 'video', 'image'}:
            message_type = 'text'

        encrypted = self.get_room(room_id).output.get('encrypted', False)  # type: ignore
        url = self.upload(
            attachment, name=filename, content_type=mime_type, encrypt=encrypted
        ).output  # type: ignore

        return {
            'url': url,
            'msgtype': 'm.' + message_type,
            'body': filename,
            'info': {
                'size': os.stat(attachment).st_size,
                'mimetype': mime_type,
            },
        }

    def _process_remote_attachment(self, attachment: str) -> dict:
        parsed_url = urlparse(attachment)
        server = parsed_url.netloc.strip('/')
        media_id = parsed_url.path.strip('/')

        response = self._loop_execute(self.client.download(server, media_id))

        content_type = response.content_type
        message_type = content_type.split('/')[0]
        if message_type not in {'audio', 'video', 'image'}:
            message_type = 'text'

        return {
            'url': attachment,
            'msgtype': 'm.' + message_type,
            'body': response.filename,
            'info': {
                'size': len(response.body),
                'mimetype': content_type,
            },
        }

    def _process_attachment(self, attachment: str, room_id: str):
        if attachment.startswith('mxc://'):
            return self._process_remote_attachment(attachment)
        return self._process_local_attachment(attachment, room_id=room_id)

    @action
    def send_message(
        self,
        room_id: str,
        message_type: str = 'text',
        body: str | None = None,
        attachment: str | None = None,
        tx_id: str | None = None,
        ignore_unverified_devices: bool = False,
    ):
        """
        Send a message to a room.

        :param room_id: Room ID.
        :param body: Message body.
        :param attachment: Path to a local file to send as an attachment, or
            URL of an existing Matrix media ID in the format
            ``mxc://<server>/<media_id>``. If the attachment is a local file,
            the file will be automatically uploaded, ``message_type`` will be
            automatically inferred from the file and the ``body`` will be
            replaced by the filename.
        :param message_type: Message type. Supported: `text`, `audio`, `video`,
            `image`. Default: `text`.
        :param tx_id: Unique transaction ID to associate to this message.
        :param ignore_unverified_devices: If true, unverified devices will be
            ignored. Otherwise, if the room is encrypted and it contains
            devices that haven't been marked as trusted, the message
            delivery may fail (default: False).
        :return: .. schema:: matrix.MatrixEventIdSchema
        """
        content = {
            'msgtype': 'm.' + message_type,
            'body': body,
        }

        if attachment:
            content.update(self._process_attachment(attachment, room_id=room_id))

        ret = self._loop_execute(
            self.client.room_send(
                message_type='m.room.message',
                room_id=room_id,
                tx_id=tx_id,
                ignore_unverified_devices=ignore_unverified_devices,
                content=content,
            )
        )

        ret = self._loop_execute(ret.transport_response.json())
        return MatrixEventIdSchema().dump(ret)

    @action
    def get_profile(self, user_id: str):
        """
        Retrieve the details about a user.

        :param user_id: User ID.
        :return: .. schema:: matrix.MatrixProfileSchema
        """
        profile = self._loop_execute(self.client.get_profile(user_id))  # type: ignore
        profile.user_id = user_id
        return MatrixProfileSchema().dump(profile)

    @action
    def get_room(self, room_id: str):
        """
        Retrieve the details about a room.

        :param room_id: room ID.
        :return: .. schema:: matrix.MatrixRoomSchema
        """
        response = self._loop_execute(self.client.room_get_state(room_id))  # type: ignore
        room_args = {'room_id': room_id, 'own_user_id': None, 'encrypted': False}
        room_params = {}

        for evt in response.events:
            if evt.get('type') == 'm.room.create':
                room_args['own_user_id'] = evt.get('content', {}).get('creator')
            elif evt.get('type') == 'm.room.encryption':
                room_args['encrypted'] = True
            elif evt.get('type') == 'm.room.name':
                room_params['name'] = evt.get('content', {}).get('name')
            elif evt.get('type') == 'm.room.topic':
                room_params['topic'] = evt.get('content', {}).get('topic')

        room = MatrixRoom(**room_args)
        for k, v in room_params.items():
            setattr(room, k, v)
        return MatrixRoomSchema().dump(room)

    @action
    def get_messages(
        self,
        room_id: str,
        start: str | None = None,
        end: str | None = None,
        backwards: bool = True,
        limit: int = 10,
    ):
        """
        Retrieve a list of messages from a room.

        :param room_id: Room ID.
        :param start: Start retrieving messages from this batch ID (default:
            latest batch returned from a call to ``sync``).
        :param end: Retrieving messages until this batch ID.
        :param backwards: Set to True if you want to retrieve messages starting
            from the most recent, in descending order (default). Otherwise, the
            first returned message will be the oldest and messages will be
            returned in ascending order.
        :param limit: Maximum number of messages to be returned (default: 10).
        :return: .. schema:: matrix.MatrixMessagesResponseSchema
        """
        response = self._loop_execute(
            self.client.room_messages(
                room_id,
                start=start,
                end=end,
                limit=limit,
                direction=(
                    MessageDirection.back if backwards else MessageDirection.front
                ),
            )
        )

        response.chunk = [m for m in response.chunk if isinstance(m, RoomMessage)]
        return MatrixMessagesResponseSchema().dump(response)

    @action
    def get_my_devices(self):
        """
        Get the list of devices associated to the current user.

        :return: .. schema:: matrix.MatrixMyDeviceSchema(many=True)
        """
        response = self._loop_execute(self.client.devices())
        return MatrixMyDeviceSchema().dump(response.devices, many=True)

    @action
    def get_device(self, device_id: str):
        """
        Get the info about a device given its ID.

        :return: .. schema:: matrix.MatrixDeviceSchema
        """
        return MatrixDeviceSchema().dump(self._get_device(device_id))

    @action
    def update_device(self, device_id: str, display_name: str | None = None):
        """
        Update information about a user's device.

        :param display_name: New display name.
        :return: .. schema:: matrix.MatrixDeviceSchema
        """
        content = {}
        if display_name:
            content['display_name'] = display_name

        self._loop_execute(self.client.update_device(device_id, content))
        return MatrixDeviceSchema().dump(self._get_device(device_id))

    @action
    def delete_devices(
        self,
        devices: Sequence[str],
        username: str | None = None,
        password: str | None = None,
    ):
        """
        Delete a list of devices from the user's authorized list and invalidate
        their access tokens.

        :param devices: List of devices that should be deleted.
        :param username: Username, if the server requires authentication upon
            device deletion.
        :param password: User password, if the server requires authentication
            upon device deletion.
        """
        auth = {}
        if username and password:
            auth = {'type': 'm.login.password', 'user': username, 'password': password}

        self._loop_execute(self.client.delete_devices([*devices], auth=auth))

    @action
    def get_joined_rooms(self):
        """
        Retrieve the rooms that the user has joined.

        :return: .. schema:: matrix.MatrixRoomSchema(many=True)
        """
        response = self._loop_execute(self.client.joined_rooms())
        return [self.get_room(room_id).output for room_id in response.rooms]  # type: ignore

    @action
    def get_room_members(self, room_id: str):
        """
        Retrieve the list of users joined into a room.

        :param room_id: The room ID.
        :return: .. schema:: matrix.MatrixMemberSchema(many=True)
        """
        response = self._loop_execute(self.client.joined_members(room_id))
        return MatrixMemberSchema().dump(response.members, many=True)

    @action
    def room_alias_to_id(self, alias: str) -> str:
        """
        Convert a room alias (in the format ``#alias:matrix.example.org``) to a
        room ID (in the format ``!aBcDeFgHiJkMnO:matrix.example.org``).

        :param alias: The room alias.
        :return: The room ID, as a string.
        """
        response = self._loop_execute(self.client.room_resolve_alias(alias))
        return response.room_id

    @action
    def add_room_alias(self, room_id: str, alias: str):
        """
        Add an alias to a room.

        :param room_id: An existing room ID.
        :param alias: The room alias.
        """
        self._loop_execute(self.client.room_put_alias(alias, room_id))

    @action
    def delete_room_alias(self, alias: str):
        """
        Delete a room alias.

        :param alias: The room alias.
        """
        self._loop_execute(self.client.room_delete_alias(alias))

    @action
    def upload_keys(self):
        """
        Synchronize the E2EE keys with the homeserver.
        """
        self._loop_execute(self.client.keys_upload())

    def _get_device(self, device_id: str) -> OlmDevice:
        device = self.client.get_device(device_id)
        assert device, f'No such device_id: {device_id}'
        return device

    @action
    def trust_device(self, device_id: str):
        """
        Mark a device as trusted.

        :param device_id: Device ID.
        """
        device = self._get_device(device_id)
        self.client.verify_device(device)

    @action
    def untrust_device(self, device_id: str):
        """
        Mark a device as untrusted.

        :param device_id: Device ID.
        """
        device = self._get_device(device_id)
        self.client.unverify_device(device)

    @action
    def mxc_to_http(self, url: str, homeserver: str | None = None) -> str:
        """
        Convert a Matrix URL (in the format ``mxc://server/media_id``) to an
        HTTP URL.

        Note that invoking this function on a URL containing encrypted content
        (i.e. a URL containing media sent to an encrypted room) will provide a
        URL that points to encrypted content. The best way to deal with
        encrypted media is by using :meth:`.download` to download the media
        locally.

        :param url: The MXC URL to be converted.
        :param homeserver: The hosting homeserver (default: the same as the URL).
        :return: The converted HTTP(s) URL.
        """
        http_url = Api.mxc_to_http(url, homeserver=homeserver)
        assert http_url, f'Could not convert URL {url}'
        return http_url

    @action
    def download(
        self,
        url: str,
        download_path: str | None = None,
        filename: str | None = None,
        allow_remote=True,
    ):
        """
        Download a file given its Matrix URL.

        Note that URLs that point to encrypted resources will be automatically
        decrypted only if they were received on a room joined by this account.

        :param url: Matrix URL, in the format ``mxc://<server>/<media_id>``.
        :param download_path: Override the default ``download_path`` (output
            directory for the downloaded file).
        :param filename: Name of the output file (default: inferred from the
            remote resource).
        :param allow_remote: Indicates to the server that it should not attempt
            to fetch the media if it is deemed remote. This is to prevent
            routing loops where the server contacts itself.
        :return: .. schema:: matrix.MatrixDownloadedFileSchema
        """
        parsed_url = urlparse(url)
        server = parsed_url.netloc.strip('/')
        media_id = parsed_url.path.strip('/')

        response = self._loop_execute(
            self.client.download(
                server, media_id, filename=filename, allow_remote=allow_remote
            )
        )

        if not download_path:
            download_path = self._download_path
        if not filename:
            filename = response.filename or media_id

        outfile = os.path.join(str(download_path), str(filename))
        pathlib.Path(download_path).mkdir(parents=True, exist_ok=True)

        with open(outfile, 'wb') as f:
            f.write(response.body)

        return MatrixDownloadedFileSchema().dump(
            {
                'url': url,
                'path': outfile,
                'size': len(response.body),
                'content_type': response.content_type,
            }
        )

    @action
    def upload(
        self,
        file: str,
        name: str | None = None,
        content_type: str | None = None,
        encrypt: bool = False,
    ) -> str:
        """
        Upload a file to the server.

        :param file: Path to the file to upload.
        :param name: Filename to be used for the remote file (default: same as
            the local file).
        :param content_type: Specify a content type for the file (default:
            inferred from the file's extension and content).
        :param encrypt: Encrypt the file (default: False).
        :return: The Matrix URL of the uploaded resource.
        """
        rs = self._loop_execute(
            self.client.upload_file(file, name, content_type, encrypt)
        )

        return rs[0].content_uri

    @action
    def create_room(
        self,
        name: str | None = None,
        alias: str | None = None,
        topic: str | None = None,
        is_public: bool = False,
        is_direct: bool = False,
        federate: bool = True,
        encrypted: bool = False,
        invite_users: Sequence[str] = (),
    ):
        """
        Create a new room on the server.

        :param name: Room name.
        :param alias: Custom alias for the canonical name. For example, if set
            to ``foo``, the alias for this room will be
            ``#foo:matrix.example.org``.
        :param topic: Room topic.
        :param is_public: Set to True if you want the room to be public and
            discoverable (default: False).
        :param is_direct: Set to True if this should be considered a direct
            room with only one user (default: False).
        :param federate: Whether you want to allow users from other servers to
            join the room (default: True).
        :param encrypted: Whether the room should be encrypted (default: False).
        :param invite_users: A list of user IDs to invite to the room.
        :return: .. schema:: matrix.MatrixRoomIdSchema
        """
        rs = self._loop_execute(
            self.client.room_create(
                name=name,
                alias=alias,
                topic=topic,
                is_direct=is_direct,
                federate=federate,
                invite=invite_users,
                visibility=(
                    RoomVisibility.public if is_public else RoomVisibility.private
                ),
                initial_state=[
                    {
                        'type': 'm.room.encryption',
                        'content': {
                            'algorithm': 'm.megolm.v1.aes-sha2',
                        },
                    }
                ]
                if encrypted
                else (),
            )
        )

        return MatrixRoomIdSchema().dump(rs)

    @action
    def invite(self, room_id: str, user_id: str):
        """
        Invite a user to a room.

        :param room_id: Room ID.
        :param user_id: User ID.
        """
        self._loop_execute(self.client.room_invite(room_id, user_id))

    @action
    def kick(self, room_id: str, user_id: str, reason: str | None = None):
        """
        Kick a user out of a room.

        :param room_id: Room ID.
        :param user_id: User ID.
        :param reason: Optional reason.
        """
        self._loop_execute(self.client.room_kick(room_id, user_id, reason))

    @action
    def ban(self, room_id: str, user_id: str, reason: str | None = None):
        """
        Ban a user from a room.

        :param room_id: Room ID.
        :param user_id: User ID.
        :param reason: Optional reason.
        """
        self._loop_execute(self.client.room_ban(room_id, user_id, reason))

    @action
    def unban(self, room_id: str, user_id: str):
        """
        Remove a user ban from a room.

        :param room_id: Room ID.
        :param user_id: User ID.
        """
        self._loop_execute(self.client.room_unban(room_id, user_id))

    @action
    def join(self, room_id: str):
        """
        Join a room.

        :param room_id: Room ID.
        """
        self._loop_execute(self.client.join(room_id))

    @action
    def leave(self, room_id: str):
        """
        Leave a joined room.

        :param room_id: Room ID.
        """
        self._loop_execute(self.client.room_leave(room_id))

    @action
    def forget(self, room_id: str):
        """
        Leave a joined room and forget its data as well as all the messages.

        If all the users leave a room, that room will be marked for deletion by
        the homeserver.

        :param room_id: Room ID.
        """
        self._loop_execute(self.client.room_forget(room_id))

    @action
    def set_display_name(self, display_name: str):
        """
        Set/change the display name for the current user.

        :param display_name: New display name.
        """
        self._loop_execute(self.client.set_displayname(display_name))

    @action
    def set_avatar(self, url: str):
        """
        Set/change the avatar URL for the current user.

        :param url: New avatar URL. It must be a valid ``mxc://`` link.
        """
        self._loop_execute(self.client.set_avatar(url))


# vim:sw=4:ts=4:et:



from threading import RLock
from typing import Optional

from platypush.config import Config
from platypush.context import get_backend, get_plugin

_app_tunnel_lock = RLock()
_app_tunnel_url: Optional[str] = None


def _get_http_port() -> int:
    http = None
    if Config.get('backend.http'):
        http = get_backend('http')

    assert http, 'The http backend is required in order to subscribe to notifications'
    return http.port


def get_or_create_ngrok_tunnel() -> str:
    """
    This method creates an ngrok tunnel for the local web application,
    useful to register public HTTPS callback URLs on the fly from plugins
    and backends.
    """
    global _app_tunnel_url
    with _app_tunnel_lock:
        if _app_tunnel_url:
            return _app_tunnel_url

        local_port = _get_http_port()

        ngrok = None
        if Config.get('ngrok'):
            ngrok = get_plugin('ngrok')

        assert ngrok, 'The ngrok plugin is required in order to subscribe to notifications'
        tunnel_response = ngrok.create_tunnel(
            resource=local_port,
            protocol='http',
            bind_tls=True,
        ).output

        _app_tunnel_url = tunnel_response.get('url')
        assert _app_tunnel_url, 'Unable to create an ngrok tunnel'
        return _app_tunnel_url



from typing import Optional, Iterable, Any

from platypush.message.event import Event


class DbusSignalEvent(Event):
    """
    Event triggered when a signal is received on the D-Bus.
    """
    def __init__(
        self, bus: str, interface: str, sender: str, path: str, signal: str,
        params: Optional[Iterable[Any]] = None, **kwargs
    ):
        """
        :param bus: Bus type (``session`` or ``system``).
        :param interface: Name of the interface associated to the signal.
        :param sender: D-Bus name of the sender of the signal.
        :param path: Path of the object associated to the signal.
        :param signal: Signal name.
        :param params: Signal payload.
        """
        super().__init__(bus=bus, interface=interface, sender=sender,
                         path=path, signal=signal, params=params, **kwargs)



import base64
import json

from platypush.backend import Backend
from platypush.message.event.nfc import NFCTagDetectedEvent, NFCTagRemovedEvent, NFCDeviceConnectedEvent, \
    NFCDeviceDisconnectedEvent


class NfcBackend(Backend):
    """
    Backend to detect NFC card events from a compatible reader.

    Triggers:

        * :class:`platypush.message.event.nfc.NFCDeviceConnectedEvent` when an NFC reader/writer is connected
        * :class:`platypush.message.event.nfc.NFCDeviceDisconnectedEvent` when an NFC reader/writer is disconnected
        * :class:`platypush.message.event.nfc.NFCTagDetectedEvent` when an NFC tag is detected
        * :class:`platypush.message.event.nfc.NFCTagRemovedEvent` when an NFC tag is removed

    Requires:

        * **nfcpy** >= 1.0 (``pip install 'nfcpy>=1.0'``)
        * **ndef** (``pip install ndef``)

    Run the following to check if your device is compatible with nfcpy and the right permissions are set::

        python -m nfc

    """

    def __init__(self, device='usb', *args, **kwargs):
        """
        :param device: Address or ID of the device to be opened. Examples:

            * `'usb:003:009'` opens device 9 on bus 3
            * `'usb:003'` opens the first available device on bus 3
            * `'usb'` opens the first available USB device (default)
        """

        super().__init__(*args, **kwargs)

        self.device_id = device
        self._clf = None

    def _get_clf(self):
        import nfc

        if not self._clf:
            self._clf = nfc.ContactlessFrontend()
            self._clf.open(self.device_id)
            self.bus.post(NFCDeviceConnectedEvent(reader=self._get_device_str()))
            self.logger.info('Initialized NFC reader backend on device {}'.format(self._get_device_str()))

        return self._clf

    def _get_device_str(self):
        return str(self._clf.device)

    def close(self):
        if self._clf:
            self._clf.close()
            self._clf = None
            self.bus.post(NFCDeviceDisconnectedEvent(reader=self._get_device_str()))

    @staticmethod
    def _parse_records(tag):
        from ndef.text import TextRecord
        from ndef.uri import UriRecord
        from ndef.smartposter import SmartposterRecord
        from ndef.deviceinfo import DeviceInformationRecord
        from ndef.wifi import WifiSimpleConfigRecord, WifiPeerToPeerRecord
        from ndef.bluetooth import BluetoothLowEnergyRecord, BluetoothEasyPairingRecord
        from ndef.signature import SignatureRecord

        records = []

        if not tag.ndef:
            return records

        for record in tag.ndef.records:
            r = {
                'record_type': record.type,
                'record_name': record.name,
            }

            if isinstance(record, TextRecord):
                try:
                    r = {
                        **r,
                        'type': 'json',
                        'value': json.loads(record.text),
                    }
                except ValueError:
                    r = {
                        **r,
                        'type': 'text',
                        'text': record.text,
                    }
            elif isinstance(record, UriRecord):
                r = {
                    **r,
                    'type': 'uri',
                    'uri': record.uri,
                    'iri': record.iri,
                }
            elif isinstance(record, SmartposterRecord):
                r = {
                    **r,
                    'type': 'smartposter',
                    **{attr: getattr(record, attr) for attr in ['resource', 'titles', 'title', 'action', 'icon',
                                                                'icons', 'resource_size', 'resource_type']},
                }
            elif isinstance(record, DeviceInformationRecord):
                r = {
                    **r,
                    'type': 'device_info',
                    **{attr: getattr(record, attr) for attr in ['vendor_name', 'model_name', 'unique_name',
                                                                'uuid_string', 'version_string']},
                }
            elif isinstance(record, WifiSimpleConfigRecord):
                r = {
                    **r,
                    'type': 'wifi_simple_config',
                    **{attr: record[attr] for attr in record.attribute_names()}
                }
            elif isinstance(record, WifiPeerToPeerRecord):
                r = {
                    **r,
                    'type': 'wifi_peer_to_peer',
                    **{attr: record[attr] for attr in record.attribute_names()}
                }
            elif isinstance(record, BluetoothEasyPairingRecord):
                r = {
                    **r,
                    'type': 'bluetooth_easy_pairing',
                    **{attr: getattr(record, attr) for attr in ['device_address', 'device_name', 'device_class']},
                }
            elif isinstance(record, BluetoothLowEnergyRecord):
                r = {
                    **r,
                    'type': 'bluetooth_low_energy',
                    **{attr: getattr(record, attr) for attr in ['device_address', 'device_name', 'role_capabilities',
                                                                'appearance', 'flags', 'security_manager_tk_value',
                                                                'secure_connections_confirmation_value',
                                                                'secure_connections_random_value']},
                }
            elif isinstance(record, SignatureRecord):
                r = {
                    **r,
                    'type': 'signature',
                    **{attr: getattr(record, attr) for attr in ['version', 'signature_type', 'hash_type', 'signature',
                                                                'signature_uri', 'certificate_format',
                                                                'certificate_store', 'certificate_uri',
                                                                'secure_connections_random_value']},
                }
            else:
                r = {
                    **r,
                    'type': 'binary',
                    'data': base64.encodebytes(record.data).decode(),
                }

            records.append(r)

        return records

    @staticmethod
    def _parse_id(tag):
        return ''.join([('%02X' % c) for c in tag.identifier])

    def _on_connect(self):
        def callback(tag):
            if not tag:
                return False

            tag_id = self._parse_id(tag)
            records = self._parse_records(tag)
            self.bus.post(NFCTagDetectedEvent(reader=self._get_device_str(), tag_id=tag_id, records=records))
            return True

        return callback

    def _on_release(self):
        def callback(tag):
            tag_id = self._parse_id(tag)
            self.bus.post(NFCTagRemovedEvent(reader=self._get_device_str(), tag_id=tag_id))

        return callback

    def run(self):
        super().run()

        while not self.should_stop():
            try:
                clf = self._get_clf()
                clf.connect(rdwr={
                    'on-connect': self._on_connect(),
                    'on-release': self._on_release(),
                })
            finally:
                self.close()


# vim:sw=4:ts=4:et:



import logging
import threading
import time

from queue import Queue, Empty
from typing import Callable, Type

from platypush.message.event import Event

logger = logging.getLogger('platypush:bus')


class Bus(object):
    """ Main local bus where the daemon will listen for new messages """

    _MSG_EXPIRY_TIMEOUT = 60.0  # Consider a message on the bus as expired after one minute without being picked up

    def __init__(self, on_message=None):
        self.bus = Queue()
        self.on_message = on_message
        self.thread_id = threading.get_ident()
        self.event_handlers = {}
        self._should_stop = threading.Event()

    def post(self, msg):
        """ Sends a message to the bus """
        self.bus.put(msg)

    def get(self):
        """ Reads one message from the bus """
        try:
            return self.bus.get(timeout=0.1)
        except Empty:
            return

    def stop(self):
        self._should_stop.set()

    def _msg_executor(self, msg):
        def event_handler(event: Event, handler: Callable[[Event], None]):
            logger.info('Triggering event handler {}'.format(handler.__name__))
            handler(event)

        def executor():
            if isinstance(msg, Event):
                if type(msg) in self.event_handlers:
                    handlers = self.event_handlers[type(msg)]
                else:
                    handlers = {*[hndl for event_type, hndl in self.event_handlers.items()
                                  if isinstance(msg, event_type)]}

                for hndl in handlers:
                    threading.Thread(target=event_handler, args=(msg, hndl))

            try:
                self.on_message(msg)
            except Exception as e:
                logger.error('Error on processing message {}'.format(msg))
                logger.exception(e)

        return executor

    def should_stop(self):
        return self._should_stop.is_set()

    def poll(self):
        """
        Reads messages from the bus until either stop event message or KeyboardInterrupt
        """
        if not self.on_message:
            logger.warning('No message handlers installed, cannot poll')
            return

        while not self.should_stop():
            msg = self.get()
            if msg is None:
                continue

            timestamp = msg.timestamp if hasattr(msg, 'timestamp') else msg.get('timestamp')
            if timestamp and time.time() - timestamp > self._MSG_EXPIRY_TIMEOUT:
                logger.debug('{} seconds old message on the bus expired, ignoring it: {}'.
                             format(int(time.time()-msg.timestamp), msg))
                continue

            threading.Thread(target=self._msg_executor(msg)).start()

        logger.info('Bus service stopped')

    def register_handler(self, event_type: Type[Event], handler: Callable[[Event], None]) -> Callable[[], None]:
        """
        Register an event handler to the bus.

        :param event_type: Event type to subscribe (event inheritance also works).
        :param handler: Event handler - a function that takes an Event object as parameter.
        :return: A function that can be called to remove the handler (no parameters required).
        """
        if event_type not in self.event_handlers:
            self.event_handlers[event_type] = set()

        self.event_handlers[event_type].add(handler)

        def unregister():
            self.unregister_handler(event_type, handler)

        return unregister

    def unregister_handler(self, event_type: Type[Event], handler: Callable[[Event], None]) -> None:
        """
        Remove an event handler.

        :param event_type: Event type.
        :param handler: Existing event handler.
        """
        if event_type not in self.event_handlers:
            return

        if handler in self.event_handlers[event_type]:
            self.event_handlers[event_type].remove(handler)

        if len(self.event_handlers[event_type]) == 0:
            del self.event_handlers[event_type]


# vim:sw=4:ts=4:et:
